"""Autocomplete helper: determine context of the token to be autocompleted"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_completion_env.ipynb.

# %% auto 0
__all__ = ['stata_lexer', 'CompletionEnv', 'Env']

# %% ../nbs/10_completion_env.ipynb 4
from .code_utils import ending_sc_delimiter
from fastcore.basics import patch_to
from enum import IntEnum
from typing import Tuple
import re

# %% ../nbs/10_completion_env.ipynb 5
from pygments import lexers
from pygments.token import Comment, Keyword, Name, Number, \
    String, Text, Operator

# %% ../nbs/10_completion_env.ipynb 7
stata_lexer = lexers.get_lexer_by_name('stata')

def _lex_tokens(code):
    return list(stata_lexer.get_tokens_unprocessed(code))

# %% ../nbs/10_completion_env.ipynb 8
def _last_token(code):
    tokens = _lex_tokens(code)
    last_tokentype = tokens[-1][1]
    tokens_to_combine = []
    for token in reversed(tokens):
        if token[1] is last_tokentype:
            tokens_to_combine.append(token)
        else:
            break
    tokens_to_combine = list(reversed(tokens_to_combine))
    return (min(tokens_to_combine, key=lambda t: t[0])[0], last_tokentype, "".join([t[2] for t in tokens_to_combine]))

# %% ../nbs/10_completion_env.ipynb 9
def _last_token_full_string(code, sc_delimiter=False):
    if not code:
        return (0, None, "")
    prefix = ""
    if sc_delimiter:
        prefix = "#delimit;\n"
        orig_code = code
        code = prefix + orig_code
    tokens = _lex_tokens(code)
    last_tokentype = tokens[-1][1]
    tokens_to_combine = []
    for token in reversed(tokens):
        if token[1] is last_tokentype:
            tokens_to_combine.append(token)
        else:
            break
    tokens_to_combine = list(reversed(tokens_to_combine))
    index = min(tokens_to_combine, key=lambda t: t[0])[0]
    value = "".join([t[2] for t in tokens_to_combine])
    while last_tokentype == String and value[0] != '"' and value[0:2] != '`"':
        tokens_to_combine = list(reversed(tokens_to_combine))
        reversed_remaining_tokens = list(reversed(_lex_tokens(code[:index])))
        for i, token in enumerate(reversed_remaining_tokens):
            if (token[1] is not String 
                and reversed_remaining_tokens[i-1][1] is String):
                break
            elif token[1] in [Comment.Single, Comment.Multiline, Comment.Special]:
                break
            tokens_to_combine.append(token)
        tokens_to_combine = list(reversed(tokens_to_combine))
        index = min(tokens_to_combine, key=lambda t: t[0])[0]
        value = "".join([t[2] for t in tokens_to_combine])
    index = index-len(prefix)
    if index < 0:
        value = value[-index:]
        index = 0
    return (index, last_tokentype, value)

# %% ../nbs/10_completion_env.ipynb 29
class CompletionEnv():
    def __init__(self):
        """"""
        self.last_word = re.compile(
            r'\W\w*?\Z', flags=re.MULTILINE).search
        
        # any non-space/"/= 'chunk' at the end of the string after the last ", =, or white space
        self.last_chunk = re.compile(
            r'[\s"=][^\s"=]*?\Z', flags=re.MULTILINE).search
        
        # Path completion
        self.path_search = re.compile(
            r'^(?P<fluff>.*")(?P<path>[^"]*)\Z').search

        # Magic completion
        self.magic_completion = re.compile(
            r'\A\*?(?P<magic>%%?\S*)\Z', flags=re.DOTALL + re.MULTILINE).match

        # Match context; this is used to determine if the line starts
        # with matrix or scalar. It also matches constructs like
        #
        #     (`=)?scalar(
        pre = (
            r'(cap(t|tu|tur|ture)?'
            r'|qui(e|et|etl|etly)?'
            r'|n(o|oi|ois|oisi|oisil|oisily)?)')
        kwargs = {'flags': re.MULTILINE}
        self.fcontext = {
            'function':
                re.compile(
                    r"(\s+|\=|`=)\s*(?P<name>\w+?)"
                    r"\([^\)]*?(?P<last_word>\w*)\Z", **kwargs).search,
        }
        self.context = {
            'line':
                re.compile(
                    r"^(?P<last_line>\s*({0}\s+)*(?P<first_word>\S+) .*?)\Z".format(pre),
                    **kwargs).search,
            'delimit_line':
                re.compile(
                    r"(?:\A|;)(?P<last_line>\s*({0}\s+)*(?P<first_word>[^\s;]+)\s[^;]*?)\Z".format(pre),
                    **kwargs).search
        }
#         self.last_line = {
#             'line':
#                 re.compile(
#                     r"^(?P<last_line>.*)\Z",
#                     **kwargs).search,
#             'delimit_line':
#                 re.compile(
#                     r"(?:\A|;)(?P<last_line>[^;]*)\Z",
#                     **kwargs).search
#         }

#         self.ends_in_a_comment = re.compile(
#             r'('
#             r'(^((\s*\*)|((.*( |\t))?\/\/)).*)'     # last line starting with '*' or containing ' //'
#             r'|(\/\*)([^\*\/]|\*(?!\/)|\/(?<!\*))*' # unfinished /* block
#             r')\Z', flags=re.MULTILINE).search

#         self.ends_in_a_string_literal = re.compile(
#             r'(\`\")' # start of a `" block
#             r'(' 
#             r'([^\"\']|\"(?!\')|\'(?<!\"))*' 
#             r'([^\"\']|\"(?!\')|\'(?<!\"))*' 
#             r'([^\"\']|\"(?!\')|\'(?<!\"))*' 
#             r')*\Z').search

# %% ../nbs/10_completion_env.ipynb 30
def _ends_in_string_literal(code, sc_delimiter=False):
    if sc_delimiter:
        code = "#delimit;\n" + code
    return _last_token(code)[1] is String

# %% ../nbs/10_completion_env.ipynb 31
def _ends_in_a_comment(code, sc_delimiter=False):
    if sc_delimiter:
        code = "#delimit;\n" + code
    return _last_token(code)[1] in [Comment.Single, Comment.Multiline, Comment.Special]

# %% ../nbs/10_completion_env.ipynb 38
@patch_to(CompletionEnv)
def _scalar_f_pos_rcomp(self, code, r2chars):
    scalar_f = False
    funcontext = self.fcontext['function'](code)
    if funcontext:
        function = funcontext.group('name')
        if function == 'scalar':
            scalar_f = True
            pos = funcontext.start('last_word') if funcontext.start('last_word') else len(code)
            rcomp = "" if (r2chars[0:1] == ")" or r2chars == " )") else ")"
    if scalar_f:
        return True, pos, rcomp
    else:
        return False, None, None

# %% ../nbs/10_completion_env.ipynb 44
@patch_to(CompletionEnv)
def _start_of_last_chunk(self, code):
    search = self.last_chunk(code)
    return search.start() + 1 if search else 0

# %% ../nbs/10_completion_env.ipynb 47
@patch_to(CompletionEnv)
def _start_of_last_word(self, code):
    search = self.last_word(code)
    return search.start() + 1 if search else 0

# %% ../nbs/10_completion_env.ipynb 50
@patch_to(CompletionEnv)
def _last_line_first_word(self, code, sc_delimiter=False):
    if sc_delimiter:
        linecontext = self.context['delimit_line'](code)
    else:
        linecontext = self.context['line'](code)
    if linecontext:
        last_line = linecontext.groupdict()['last_line']
        first_word = linecontext.groupdict()['first_word']
        return last_line, first_word
    else:
        return None, None

# %% ../nbs/10_completion_env.ipynb 56
class Env(IntEnum):
    NONE = -9      # no suggestions
    MAGIC = -1     # magics, e.g., %x*
    GENERAL = 0    # varlist and/or file path
    LOCAL = 1      # `x* completed with `x*'
    GLOBAL = 2     # $x* completed with $x* or ${x* completed with ${x*}
    SCALAR = 4     # scalar .* x* completed with x* or scalar(x* completed with scalar(x*)
    MATRIX = 6     # matrix .* x* completed with x*
    SCALAR_VAR = 7 # scalars and varlist, scalar .* = x* completed with x*
    MATRIX_VAR = 8 # matrices and varlist, matrix .* = x* completed with x*
    MATA = 9       # inline or in mata environment
    STRING = 10    # file path

# %% ../nbs/10_completion_env.ipynb 57
@patch_to(CompletionEnv)
def get_env(self, 
            code: str, # Right-truncated to cursor position
            r2chars: str, # The two characters immediately after `code`, used to accurately determine rcomp
            sc_delimiter,
           ) -> Tuple[Env, int, str, str]:
    """Returns completions environment
    
    Returns
    -------
    env : Env    
    pos : int
        Where the completions start. This is set to the start of the word to be completed.
    out_chunk : str
        Word to match.
    rcomp : str
        How to finish the completion (defaulting to nothing):
        locals: '
        globals (if start with ${): }
        scalars: )
        scalars (if start with `): )'
    """
    rcomp = ""
    
    lcode = code.lstrip()
    if self.magic_completion(lcode):
        pos = code.find("%")
        env = Env.MAGIC
        return env, pos, code[pos:], rcomp
    
    sc_delimiter = ending_sc_delimiter(code, sc_delimiter)
    env = Env.GENERAL   
    
    pos = self._start_of_last_word(code)

    if _ends_in_a_comment(code, sc_delimiter):
        return env, pos, code[pos:], rcomp

    last_token_index, last_token_type, last_token_value = _last_token_full_string(code, sc_delimiter)
    
    if last_token_type is String:
        if (not _ends_in_string_literal(code + " ", sc_delimiter)
            or not (last_token_value.startswith('"')
                    or last_token_value.startswith('`"'))):
            return Env.NONE, len(code)-1, rcomp
        if last_token_value.startswith('"'):
            opening_marker_length = 1
            rcomp = "" if r2chars[0:1] == '"' else '"'
        elif last_token_value.startswith('`"'):
            opening_marker_length = 2
            rcomp = "" if r2chars[0:2] == "\"'" else "\"'"
        pos = last_token_index + opening_marker_length
        env = Env.STRING
    else:
        # Figure out if this is a local or global; env = 0 (default)
        # will suggest variables in memory.
        cpos = self._start_of_last_chunk(code) # last "chunk" delimited by white space, a double-quote, or =.
        chunk = code[cpos:]
        lfind = chunk.rfind('`')
        gfind = chunk.rfind('$')
        path_chars = any(x in chunk for x in ['/', '\\', '~'])

        if lfind >= 0 and (lfind > gfind):
            pos = cpos + lfind + 1
            env = Env.LOCAL
            rcomp = "" if r2chars[0:1] == "'" else "'"
        elif gfind >= 0 and not path_chars:
            bfind = chunk.rfind('{')
            if bfind >= 0 and (bfind == gfind+1):
                pos = cpos + bfind + 1
                env = Env.GLOBAL
                rcomp = "" if r2chars[0:1] == "}" else "}"
            else:
                env = Env.GLOBAL
                pos = cpos + gfind + 1    
    
    if pos == 0:
        env = Env.NONE # to-do: auto-complete commands here
    else:
        # Figure out if current statement is a matrix or scalar
        # statement. If so, will add them to completions list.
        last_line, first_word = self._last_line_first_word(code, sc_delimiter)
        if first_word:
            equals_present = (last_line.find('=') > 0)
            if re.match(r'^sca(lar|la|l)?$', first_word): #.strip()
                env = Env.SCALAR_VAR if equals_present else Env.SCALAR
            elif re.match(r'^mat(rix|ri|r)?$', first_word): #.strip()
                env = Env.MATRIX_VAR if equals_present else Env.MATRIX

        # Constructs of the form scalar(x<tab> will be filled only
        # with scalars. This can be preceded by = or `=
        if env in [Env.GENERAL, Env.STRING]:
            scalar_f, new_pos, new_rcomp = self._scalar_f_pos_rcomp(code, r2chars)
            if scalar_f:
                env = Env.SCALAR
                pos = new_pos
                rcomp = new_rcomp

    out_chunk = code[pos:]
    return env, pos, out_chunk, rcomp



"""Autocomplete functionality"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/11_completions.ipynb.

# %% auto 0
__all__ = ['relevant_suggestion_keys', 'CompletionsManager']

# %% ../nbs/11_completions.ipynb 4
from .stata import get_global, pwd
from .stata_session import StataSession
from .magics import StataMagics
from .completion_env import CompletionEnv, Env
from fastcore.basics import patch_to
import os
import re
import platform

# %% ../nbs/11_completions.ipynb 5
class CompletionsManager():
    def __init__(self, stata_session: StataSession):
        """"""
        self.stata_session = stata_session
        self.available_magics = list(StataMagics.available_magics.keys())
        self.env_helper = CompletionEnv()

# %% ../nbs/11_completions.ipynb 8
@patch_to(CompletionsManager)
def get_globals(self):
    if self.stata_session.suggestions:
        return {k: get_global(k) for k in self.stata_session.suggestions['globals']}
    else:
        return {}

# %% ../nbs/11_completions.ipynb 9
@patch_to(CompletionsManager)
def get_file_paths(self, chunk):
    """Get file paths based on chunk
    Args:
        chunk (str): chunk of text after last space. Doesn't include string
            punctuation characters
    Returns:
        (List[str]): folders and files at that location
    """
    # If local exists, return empty list
    if re.search(r'[`\']', chunk):
        return []

    # Define directory separator
    dir_sep = '/'
    if platform.system() == 'Windows':
        if '/' not in chunk:
            dir_sep = '\\'

    # Get directory without ending file, and without / or \
    if any(x in chunk for x in ['/', '\\']):
        ind = max(chunk.rfind('/'), chunk.rfind('\\'))
        user_folder = chunk[:ind + 1]
        user_starts = chunk[ind + 1:]

        # Replace multiple consecutive / with a single /
        user_folder = re.sub(r'/+', '/', user_folder)
        user_folder = re.sub(r'\\+', r'\\', user_folder)

    else:
        user_folder = ''
        user_starts = chunk

    # Replace globals with their values
    globals_re = r'\$\{?((?![0-9_])\w{1,32})\}?'
    try:
        folder = re.sub(
            globals_re, 
            lambda x: self.get_globals()[x.group(1)], 
            user_folder
        )
    except KeyError:
        # If the global doesn't exist (aka it hasn't been defined in 
        # the Stata environment yet), then there are no paths to check
        return []

    # Use Stata's relative path
    abspath = re.search(r'^([/~]|[a-zA-Z]:)', folder)
    if not abspath:
        folder = pwd() + '/' + folder

    try:
        top_dir, dirs, files = next(os.walk(os.path.expanduser(folder)))
        results = [x + dir_sep for x in dirs] + files
        results = [
            user_folder + x for x in results if not x.startswith('.')
            and re.match(re.escape(user_starts), x, re.I)]

    except StopIteration:
        results = []

    return sorted(results)

# %% ../nbs/11_completions.ipynb 13
relevant_suggestion_keys = {
    Env.NONE: [],
    Env.GENERAL: ['varlist', 'scalars'],
    Env.LOCAL: ['locals'],
    Env.GLOBAL: ['globals'],
    Env.SCALAR: ['scalars'],
    Env.MATRIX: ['matrices'],
    Env.SCALAR_VAR: ['scalars', 'varlist'],
    Env.MATRIX_VAR: ['matrices', 'varlist'],
    Env.STRING: [],
}

@patch_to(CompletionsManager)
def get(self, starts, env, rcomp):
    """Return environment-aware completions list."""
    if env is Env.MAGIC:
        candidate_suggestions = self.available_magics
    else:
        candidate_suggestions = [suggestion
                                 for key in relevant_suggestion_keys[env]
                                 for suggestion in self.stata_session.suggestions[key]]
    relevant_suggestions = [candidate + rcomp 
                            for candidate in candidate_suggestions
                            if candidate.startswith(starts)]
    if env in [Env.GENERAL, Env.STRING]:
        relevant_suggestions += self.get_file_paths(starts)
    return relevant_suggestions

#     elif env == 9:
#         if len(starts) > 1:
#             builtins = [
#                 var for var in mata_builtins if var.startswith(starts)]
#         else:
#             builtins = []

#         if re.search(r'[/\\]', starts):
#             paths = self.get_file_paths(starts)
#         else:
#             paths = []

#         return [
#             var for var in self.stata_session.suggestions['mata']
#             if var.startswith(starts)] + builtins + paths

# %% ../nbs/11_completions.ipynb 14
@patch_to(CompletionsManager)
def do(self, code, cursor_pos):
    if self.stata_session.suggestions is None:
        self.stata_session.refresh_suggestions()
    env, pos, chunk, rcomp = self.env_helper.get_env(
        code[:cursor_pos], 
        code[cursor_pos:(cursor_pos + 2)],
        self.stata_session.sc_delimiter,
    )
    return pos, cursor_pos, self.get(chunk, env, rcomp)



"""IPythonKernel based on pystata"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/14_kernel.ipynb.

# %% auto 0
__all__ = ['PyStataKernel', 'print_stata_error']

# %% ../nbs/14_kernel.ipynb 4
from .config import Config
from .misc_utils import print_red
from .stata_more import user_expression
from .inspect import get_inspect
from .stata_session import StataSession
from .completions import CompletionsManager
from .cell import Cell
import nbstata # for __version__
from fastcore.basics import patch_to
from ipykernel.ipkernel import IPythonKernel

# %% ../nbs/14_kernel.ipynb 6
class PyStataKernel(IPythonKernel):
    """A jupyter kernel based on pystata"""
    implementation = 'nbstata'
    implementation_version = nbstata.__version__
    language_info = {
        'name': 'stata',
        'version': '17',
        'mimetype': 'text/x-stata',
        'file_extension': '.do',
    }
    banner = "nbstata: a Jupyter kernel for Stata based on pystata"
    help_links = [
        {
            "text": "Stata Documentation",
            "url": "https://www.stata.com/features/documentation/",
        },
        {
            "text": "nbstata Help",
            "url": "https://hugetim.github.io/nbstata/",
        },
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.stata_ready = False
        self.ipydatagrid_height_set = False
        self.shell.execution_count = 0
        self.inspect_output = "Stata not yet initialized."
        self.nbstata_config = Config()
        self.stata_session = StataSession()
        self.completions = CompletionsManager(self.stata_session)
        self.inspect_output = ""

# %% ../nbs/14_kernel.ipynb 8
@patch_to(PyStataKernel)
def init_session(self):
    self.nbstata_config.process_config_file()
    self.nbstata_config.init_stata()
    self.stata_ready = True

# %% ../nbs/14_kernel.ipynb 9
def _stata_error_reply(ename, evalue, execution_count=None):
    reply_content = {
        'status': "error",
        "traceback": [],
        "ename": ename,
        "evalue": evalue,
    }
    if execution_count is not None:
        reply_content['execution_count'] = execution_count
    return reply_content

# %% ../nbs/14_kernel.ipynb 10
_missing_stata_message = (
    "pystata path not found\n"
    "A Stata 17+ installation is required to use the nbstata Stata kernel. "
    "If you already have Stata 17+ installed, "
    "please specify its path in your configuration file."
)

# %% ../nbs/14_kernel.ipynb 12
def _handle_stata_import_error(err, silent, execution_count):
    if not silent:
        print_red(f"ModuleNotFoundError: {_missing_stata_message}")
    return _stata_error_reply(
        ename = "ModuleNotFoundError", 
        evalue = _missing_stata_message, 
        execution_count = execution_count,
    )

# %% ../nbs/14_kernel.ipynb 13
def _handle_stata_init_error(err, silent, execution_count):
    reply_content = _stata_error_reply(
        ename = "Stata init error", 
        evalue = str(err), 
        execution_count = execution_count,
    )
    if not silent:
        print_red(reply_content['evalue'])
    return reply_content

# %% ../nbs/14_kernel.ipynb 17
def print_stata_error(text):
    lines = text.splitlines()
    if len(lines) >= 2 and lines[-2] == lines[-1]:
        lines.pop(-1) # remove duplicate error code glitch in pystata.stata.run multi-line (ex. below)
    if len(lines) > 2:
        print("\n".join(lines[:-2]))
    print_red("\n".join(lines[-2:]))

# %% ../nbs/14_kernel.ipynb 23
def _handle_stata_error(err, silent=False, execution_count=None):
    reply_content = _stata_error_reply(
        ename = "Stata error", 
        evalue = str(err), 
        execution_count = execution_count,
    )
    if not silent:
        print_stata_error(reply_content['evalue'])
    return reply_content

# %% ../nbs/14_kernel.ipynb 24
def _format_user_obj(user_expr_output):
    return dict(
        status='ok',
        data={'text/plain': user_expr_output},
        metadata={},
    )

# %% ../nbs/14_kernel.ipynb 25
def _user_expressions(expressions):
    results = {}
    for key, expr in expressions.items():
        try:
            value = _format_user_obj(user_expression(expr))
        except Exception as err:
            value = _stata_error_reply(
                ename = "Stata user expression error",
                evalue = str(err)
            )
            print_red(value['evalue'])
        results[key] = value
    return results

# %% ../nbs/14_kernel.ipynb 29
@patch_to(PyStataKernel)
def post_do_hook(self):
    self.inspect_output = ""

# %% ../nbs/14_kernel.ipynb 30
@patch_to(PyStataKernel)
def do_execute(self, code, silent,
               store_history=True, user_expressions=None, allow_stdin=False):
    """Execute Stata code cell"""
    if not self.stata_ready:
        try:
            self.init_session() # do this here so config error messages displayed in notebook
        except OSError as err:
            return _handle_stata_init_error(err, silent, self.execution_count)
        except ModuleNotFoundError as err: # this should almost always be preempted by OSErrors now
            return _handle_stata_import_error(err, silent, self.execution_count)
        
    self.shell.execution_count += 1
    code_cell = Cell(self, code, silent)
    try:
        code_cell.run()
    except SystemError as err:
        return _handle_stata_error(err, silent, self.execution_count)
    self.post_do_hook()
    return {
        'status': "ok",
        'execution_count': self.execution_count,
        'payload': [],
        'user_expressions': _user_expressions(user_expressions or {}),
    }

# %% ../nbs/14_kernel.ipynb 32
@patch_to(PyStataKernel)
def do_inspect(self, code, cursor_pos, detail_level=0, omit_sections=()):
    """Display Stata 'describe' output (regardless of cursor position)"""
    if self.stata_ready:
        if not self.inspect_output:
            self.inspect_output = get_inspect(code, cursor_pos, detail_level, omit_sections)
        data = {'text/plain': self.inspect_output}
    else:
        data = {}
    return {"status": "ok", "data": data, "metadata": {}, "found": True}

# %% ../nbs/14_kernel.ipynb 33
@patch_to(PyStataKernel)
def do_complete(self, code, cursor_pos):
    """Provide context-aware tab-autocomplete suggestions"""
    if self.stata_ready:
        cursor_start, cursor_end, matches = self.completions.do(
            code,
            cursor_pos,
        )
    else:
        cursor_start = cursor_end = cursor_pos
        matches = []
    return {
        'status': "ok",
        'cursor_start': cursor_start,
        'cursor_end': cursor_end,
        'metadata': {},
        'matches': matches,
    }

# %% ../nbs/14_kernel.ipynb 34
@patch_to(PyStataKernel)
def do_is_complete(self, code):
    """Overrides IPythonKernel with kernelbase default"""
    return {"status": "unknown"}

# %% ../nbs/14_kernel.ipynb 35
@patch_to(PyStataKernel)
def do_history(
    self,
    hist_access_type,
    output,
    raw,
    session=None,
    start=None,
    stop=None,
    n=None,
    pattern=None,
    unique=False,
):
    """Overrides IPythonKernel with kernelbase default"""
    return {"status": "ok", "history": []}



"""Simple wrappers for `pystata`/`sfi` functionality"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_stata.ipynb.

# %% auto 0
__all__ = ['get_local', 'set_local', 'get_global', 'get_scalar', 'stata_formatted', 'variable_names', 'drop_var', 'obs_count',
           'pwd', 'macro_expand', 'run_direct', 'run_single']

# %% ../nbs/02_stata.ipynb 5
from .misc_utils import print_red
from contextlib import redirect_stdout
from io import StringIO

# %% ../nbs/02_stata.ipynb 9
def get_local(name):
    import sfi
    return sfi.Macro.getLocal(name)

# %% ../nbs/02_stata.ipynb 11
def set_local(name, value):
    import sfi
    return sfi.Macro.setLocal(name, value)

# %% ../nbs/02_stata.ipynb 13
def get_global(name):
    import sfi
    return sfi.Macro.getGlobal(name)

# %% ../nbs/02_stata.ipynb 15
def get_scalar(name):
    import sfi
    return sfi.Scalar.getValue(name)

# %% ../nbs/02_stata.ipynb 17
def stata_formatted(value, s_format):
    import sfi
    return sfi.SFIToolkit.formatValue(value, s_format)

# %% ../nbs/02_stata.ipynb 19
def variable_names():
    from sfi import Data
    return [Data.getVarName(i) for i in range(Data.getVarCount())]

# %% ../nbs/02_stata.ipynb 23
def drop_var(name):
    import sfi
    sfi.Data.dropVar(name)

# %% ../nbs/02_stata.ipynb 26
def obs_count():
    """Count the number of observations"""
    import sfi
    return sfi.Data.getObsTotal()

# %% ../nbs/02_stata.ipynb 29
def pwd():
    from sfi import SFIToolkit
    return SFIToolkit.getWorkingDir()

# %% ../nbs/02_stata.ipynb 32
def macro_expand(s):
    from sfi import SFIToolkit
    return SFIToolkit.macroExpand(s)

# %% ../nbs/02_stata.ipynb 35
def run_direct(cmds, quietly=False, echo=False, inline=True):
    import pystata
    return pystata.stata.run(cmds, quietly, echo, inline)

# %% ../nbs/02_stata.ipynb 43
def run_single(cmd, echo=False):
    import sfi
    try:
        sfi.SFIToolkit.stata(cmd, echo)
    except Exception as e:
        with redirect_stdout(StringIO()) as diverted:
            sfi.SFIToolkit.stata("", echo)
        raise SyntaxError(diverted.getvalue())



"""IPython magics for nbstata"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/09_magics.ipynb.

# %% auto 0
__all__ = ['print_kernel', 'StataMagics', 'Frame']

# %% ../nbs/09_magics.ipynb 3
from .config import Config
from .misc_utils import print_red
from .stata import obs_count, macro_expand, get_global
from .stata_session import warn_re_unclosed_comment_block_if_needed
import nbstata.browse as browse
from fastcore.basics import patch_to
import re
import urllib
from pkg_resources import resource_filename
from bs4 import BeautifulSoup as bs
import configparser

# %% ../nbs/09_magics.ipynb 4
def print_kernel(msg, kernel):
    msg = re.sub(r'$', r'\r\n', msg, flags=re.MULTILINE)
    msg = re.sub(r'[\r\n]{1,2}[\r\n]{1,2}', r'\r\n', msg, flags=re.MULTILINE)
    stream_content = {'text': msg, 'name': 'stdout'}
    kernel.send_response(kernel.iopub_socket, 'stream', stream_content)

# %% ../nbs/09_magics.ipynb 5
def _construct_abbrev_dict():
    def _all_abbrevs(full_command, shortest_abbrev):
        for j in range(len(shortest_abbrev), len(full_command)):
            yield full_command[0:j]
    out = {}
    abbrev_list = [
        ('browse', 'br'),
        ('frbrowse', 'frbr'),
        ('help', 'h'),
        ('quietly', 'q'),
    ]
    for full_command, shortest_abbrev in abbrev_list:
        out.update(
            {abbrev: full_command
             for abbrev in _all_abbrevs(full_command, shortest_abbrev)}
        )
    return out

# %% ../nbs/09_magics.ipynb 7
class StataMagics():
    """Class for handling magics"""
    magic_regex = re.compile(
        r'\A\*?(?P<magic>%%?\w+?)(?P<code>[\s,]+.*?)?\Z', flags=re.DOTALL + re.MULTILINE)

    # Format: magic_name: help_content
    available_magics = {
        '%browse': '{} [-h] [varlist] [if] [in] [, nolabel noformat]',
        '%head': '{} [-h] [N] [varlist] [if] [, nolabel noformat]',
        '%tail': '{} [-h] [N] [varlist] [if] [, nolabel noformat]',
        '%frbrowse': '{} [-h] framename: [varlist] [if] [in] [, nolabel noformat]',
        '%frhead': '{} [-h] framename: [N] [varlist] [if] [, nolabel noformat]',
        '%frtail': '{} [-h] framename: [N] [varlist] [if] [, nolabel noformat]',
        '%locals': '',
        '%delimit': '',
        '%help': '{} [-h] command_or_topic_name',
        '%set': '{} [-h] key = value',
        '%%set': '{} [-h]\nkey1 = value1\n[key2 = value2]\n[...]',
        '%status': '',
        '%%quietly': '',
        '%%noecho': '',
        '%%echo': '',
    }
    
    abbrev_dict = _construct_abbrev_dict()
    
    csshelp_default = resource_filename(
        'nbstata', 'css/_StataKernelHelpDefault.css'
    )

    def magic_quietly(self, code, kernel, cell):
        """Suppress all display for the current cell."""
        cell.quietly = True
        return code

    def magic_noecho(self, code, kernel, cell):
        """Suppress echo for the current cell."""
        cell.noecho = True
        cell.echo = False
        return code
    
    def magic_echo(self, code, kernel, cell):
        """Suppress echo for the current cell."""
        cell.noecho = False
        cell.echo = True
        return code
    
    def magic_delimit(self, code, kernel, cell):
        delim = ';' if kernel.stata_session.sc_delimiter else 'cr'
        print_kernel(f'Current Stata command delimiter: {delim}', kernel)
        return ''
    
    def magic_status(self, code, kernel, cell):
        kernel.nbstata_config.display_status()
        return ''

# %% ../nbs/09_magics.ipynb 8
@patch_to(StataMagics)
def _unabbrev_magic_name(self, raw_name):
    last_percent = raw_name.rfind('%')
    percent_part = raw_name[:last_percent+1]
    raw_name_part = raw_name[last_percent+1:]
    if raw_name_part in self.abbrev_dict:
        name_part = self.abbrev_dict[raw_name_part]
    else:
        name_part = raw_name_part
    return percent_part + name_part

# %% ../nbs/09_magics.ipynb 11
def _parse_magic_name_code(match):
    v = match.groupdict()
    for k in v:
        v[k] = v[k] if v[k] is not None else ''                
    name = v['magic'].strip()
    code = v['code'].strip()
    return name, code

# %% ../nbs/09_magics.ipynb 12
@patch_to(StataMagics)
def _parse_code_for_magic(self, code):
    match = self.magic_regex.match(code.strip())
    if match:
        raw_name, mcode = _parse_magic_name_code(match)
        name = self._unabbrev_magic_name(raw_name)
        if name in {'%quietly', '%noecho', '%echo'}:
            print_red(
                f"Warning: The correct syntax for a cell magic is '%{name}', not '{name}'. "
                "In v1.0, nbstata may trigger an error instead of just a warning."
            )
            name = '%' + name
        elif name == "%set" and len(code.splitlines()) > 1:
            print_red(
                f"Warning: The correct syntax for the multi-line 'set' magic is '%{name}', not '{name}'. "
                "In v1.0, nbstata may trigger an error instead of just a warning."
            )
            name = '%' + name
        elif name == "%%set" and len(code.splitlines()) == 1:
            print_red(
                f"Warning: The correct syntax for the single-line 'set' magic is '%set', not '{name}'. "
                "In v1.0, nbstata may trigger an error instead of just a warning."
            )
            name = '%set'
        elif name not in self.available_magics:
            raise ValueError(f"Unknown magic {name}.")
        return name, mcode
    else:
        return None, code

# %% ../nbs/09_magics.ipynb 19
@patch_to(StataMagics)
def _do_magic(self, name, code, kernel, cell):
    if code.startswith('-h') or code.startswith('--help') or (name == "%help" and (not code or code.isspace())):
        print_kernel(self.available_magics[name].format(name), kernel)
        return ''
    else:
        return getattr(self, "magic_" + name.lstrip('%'))(code, kernel, cell)

# %% ../nbs/09_magics.ipynb 21
@patch_to(StataMagics)
def magic(self, code, kernel, cell):
    try:
        if not kernel.ipydatagrid_height_set:
            browse.set_ipydatagrid_height()
            kernel.ipydatagrid_height_set = True
        name, code = self._parse_code_for_magic(code)
    except ValueError as e:
        print_kernel(str(e), kernel)
    else:
        if name:
            code = self._do_magic(name, code, kernel, cell)
    return code        

# %% ../nbs/09_magics.ipynb 22
def _formatted_local_list(local_dict):
    std_len = 14
    str_reps = []
    for n in local_dict:
        if len(n) <= std_len:
            str_reps.append(f"{n}:{' '*(std_len-len(n))} {local_dict[n]}")
        else:
            str_reps.append(f"{n}:\n{' '*std_len}  {local_dict[n]}")
    return "\n".join(str_reps)

# %% ../nbs/09_magics.ipynb 25
@patch_to(StataMagics)
def magic_locals(self, code, kernel, cell):
    local_dict = kernel.stata_session.get_local_dict()
    print_kernel(_formatted_local_list(local_dict), kernel)
    return ''

# %% ../nbs/09_magics.ipynb 27
def _get_new_settings(code):
    parser = configparser.ConfigParser(
        empty_lines_in_values=False,
        comment_prefixes=('*','//', '/*'), # '/*': to not cause error when commenting out for Stata purposes only
        inline_comment_prefixes=('//',),
    )
    parser.read_string("[set]\n" + code.strip(), source="set")
    return dict(parser.items('set'))

# %% ../nbs/09_magics.ipynb 31
def _clean_error_message(err_str):
    return (err_str
            .replace("While reading from 'set'", "")
            .replace(" in section 'set' already exists", " already set above")
            .replace("Source contains ", "")
            .replace(" 'set'\n", "\n")
           )

# %% ../nbs/09_magics.ipynb 32
def _process_new_settings(settings, kernel):
    kernel.nbstata_config.update(settings)
    kernel.nbstata_config.update_graph_config()

# %% ../nbs/09_magics.ipynb 33
@patch_to(StataMagics)
def magic_set(self, code, kernel, cell):
    try:
        settings = _get_new_settings(code)
    except configparser.Error as err:
        print_red(f"set error:\n    {_clean_error_message(str(err))}")
    else:
        _process_new_settings(settings, kernel)
        warn_re_unclosed_comment_block_if_needed(code)

# %% ../nbs/09_magics.ipynb 38
@patch_to(StataMagics)
def magic_browse(self, code, kernel, cell):
    """Display data interactively."""
    try:
        expanded_code = macro_expand(code)
        params = browse.browse_df_params(
            expanded_code, obs_count(), kernel.nbstata_config.env['missing'],
        )
        sformat = params[-1]
        df = browse.get_df(*params)
        browse.display_df_as_ipydatagrid(df)
    except Exception as e:
        print_kernel(f"browse failed.\r\n{e}", kernel)
    return ''

# %% ../nbs/09_magics.ipynb 39
class Frame():
    """Class for generating Stata select_var for getAsDict"""
    def __init__(self, framename):
        self.original_framename = get_global('c(frame)')
        self.framename = framename
            
    def __enter__(self):
        import sfi
        try:
            frame = sfi.Frame.connect(self.framename)
        except sfi.FrameError:
            raise ValueError(f"frame {self.framename} not found")
        else:
            frame.changeToCWF()
    
    def __exit__(self, exc_type, exc_value, exc_tb):
        import sfi
        orig_frame = sfi.Frame.connect(self.original_framename)
        orig_frame.changeToCWF()

# %% ../nbs/09_magics.ipynb 40
def _parse_frame_prefix(code):
    pattern = re.compile(
        r'\A(?P<frame>\w+)[ \t]*(?:\:[ \t]*(?P<code>.*?))?\Z', flags=re.DOTALL)
    match = pattern.match(code)
    if not match:
        raise ValueError("invalid syntax: missing framename or colon?")
    v = match.groupdict()
    for k in v:
        v[k] = v[k] if v[k] is not None else ''                
    framename = v['frame'].strip()
    main_code = v['code'].strip()
    return framename, main_code

# %% ../nbs/09_magics.ipynb 42
@patch_to(StataMagics)
def magic_frbrowse(self, code, kernel, cell):
    """Display frame interactively."""
    try:
        framename, main_code = _parse_frame_prefix(code)
        with Frame(framename):
            self.magic_browse(main_code, kernel, cell)
    except Exception as e:
        print_kernel(f"frbrowse failed.\r\n{e}", kernel)
    return ''

# %% ../nbs/09_magics.ipynb 46
def _get_html_data(df):
    html = df.convert_dtypes().to_html(notebook=True)
    return {'text/html': html}

# %% ../nbs/09_magics.ipynb 47
@patch_to(StataMagics)
def _headtail_html(self, df, kernel):
    content = {
        'data': _get_html_data(df),
        'metadata': {},
    }
    kernel.send_response(kernel.iopub_socket, 'display_data', content)

# %% ../nbs/09_magics.ipynb 48
@patch_to(StataMagics)
def _magic_headtail(self, code, kernel, cell, tail=False):
    try:
        expanded_code = macro_expand(code)
        df = browse.headtail_get_df(*browse.headtail_df_params(
            expanded_code, obs_count(), kernel.nbstata_config.env['missing'], tail=tail
        ))
        self._headtail_html(df, kernel)
    except Exception as e:
        print_kernel(f"{'tail' if tail else 'head'} failed.\r\n{e}", kernel)
    return ''

# %% ../nbs/09_magics.ipynb 49
@patch_to(StataMagics)
def magic_head(self, code, kernel, cell):
    """Display data in a nicely-formatted table."""
    return self._magic_headtail(code, kernel, cell, tail=False)

# %% ../nbs/09_magics.ipynb 50
@patch_to(StataMagics)
def magic_frhead(self, code, kernel, cell):
    """Display data in a nicely-formatted table."""
    return self._magic_frheadtail(code, kernel, cell, tail=False)

# %% ../nbs/09_magics.ipynb 51
@patch_to(StataMagics)
def magic_tail(self, code, kernel, cell):
    """Display data in a nicely-formatted table."""
    return self._magic_headtail(code, kernel, cell, tail=True)

# %% ../nbs/09_magics.ipynb 52
@patch_to(StataMagics)
def magic_frtail(self, code, kernel, cell):
    """Display data in a nicely-formatted table."""
    return self._magic_frheadtail(code, kernel, cell, tail=True)

# %% ../nbs/09_magics.ipynb 53
@patch_to(StataMagics)
def _magic_frheadtail(self, code, kernel, cell, tail):
    """Display frame interactively."""
    try:
        framename, main_code = _parse_frame_prefix(code)
        with Frame(framename):
            self._magic_headtail(main_code, kernel, cell, tail)
    except Exception as e:
        print_kernel(f"{'tail' if tail else 'head'} failed.\r\n{e}", kernel)
    return ''

# %% ../nbs/09_magics.ipynb 55
@patch_to(StataMagics)
def _get_help_html(self, code):
    html_base = "https://www.stata.com"
    html_help = urllib.parse.urljoin(html_base, "help.cgi?{}")
    reply = urllib.request.urlopen(html_help.format(code))
    html = reply.read().decode("utf-8")

    # Remove excessive extra lines (Note css: "white-space: pre-wrap")
    edited_html = html.replace("<p>\n", "<p>")
    soup = bs(edited_html, 'html.parser')

    # Set root for links to https://www.stata.com
    for a in soup.find_all('a', href=True):
        href = a.get('href')
        match = re.search(r'{}(.*?)#'.format(code), href)
        if match:
            hrelative = href.find('#')
            a['href'] = href[hrelative:]
        elif not href.startswith('http'):
            link = a['href']
            match = re.search(r'/help.cgi\?(.+)$', link)
            # URL encode bad characters like %
            if match:
                link = '/help.cgi?'
                link += urllib.parse.quote_plus(match.group(1))
            a['href'] = urllib.parse.urljoin(html_base, link)
            a['target'] = '_blank'

    # Remove header 'Stata 15 help for ...'
    stata_header = soup.find('h2')
    if stata_header:
        stata_header.decompose()

    # Remove Stata help menu
    soup.find('div', id='menu').decompose()

    # Remove Copyright notice
    copyright = soup.find(string=re.compile(r".*Copyright.*", flags=re.DOTALL))
    copyright.find_parent("table").decompose()

    # Remove last hrule
    soup.find_all('hr')[-1].decompose()
    
    # Remove last br
    soup.find_all('br')[-1].decompose()
    
    # Remove last empty paragraph, empty space
    empty_paragraphs = soup.find_all('p', string="")
    if str(empty_paragraphs[-1]) == "<p></p>":
        empty_paragraphs[-1].decompose()

    # Set all the backgrounds to transparent
    for color in ['#ffffff', '#FFFFFF']:
        for bg in ['bgcolor', 'background', 'background-color']:
            for tag in soup.find_all(attrs={bg: color}):
                if tag.get(bg):
                    tag[bg] = 'transparent'

    # Set html
    css = soup.find('style', {'type': 'text/css'})
    with open(self.csshelp_default, 'r') as default:
        css.string = default.read()

    return str(soup)

# %% ../nbs/09_magics.ipynb 56
@patch_to(StataMagics)
def magic_help(self, code, kernel, cell):
    """Show help file from stata.com/help.cgi?\{\}"""
    try:
        html_help = self._get_help_html(code)
    except Exception as e: # original: (urllib.error.HTTPError, urllib.error.URLError)
        msg = "Failed to fetch HTML help.\r\n{0}"
        print_kernel(msg.format(e), kernel)
    else:
        fallback = 'This front-end cannot display HTML help.'
        resp = {
            'data': {
                'text/html': html_help,
                'text/plain': fallback},
            'metadata': {}}
        kernel.send_response(kernel.iopub_socket, 'display_data', resp)
    return ''



"""Stata-related helper functions with no Jupyter or pystata dependence"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_code_utils.ipynb.

# %% auto 0
__all__ = ['stata_lexer', 'delimit_regex', 'multi_regex', 'pre', 'kwargs', 'local_def_in', 'preserve_restore_in',
           'remove_comments', 'ends_in_comment_block', 'valid_single_line_code', 'ending_sc_delimiter',
           'standardize_code', 'ending_code_version', 'is_start_of_program_block', 'break_out_prog_blocks']

# %% ../nbs/04_code_utils.ipynb 4
import re
from decimal import Decimal

# %% ../nbs/04_code_utils.ipynb 5
from pygments import lexers
from pygments.token import Comment

# %% ../nbs/04_code_utils.ipynb 8
stata_lexer = lexers.get_lexer_by_name('stata')

def _lex_tokens(code):
    return stata_lexer.get_tokens_unprocessed(code)

# %% ../nbs/04_code_utils.ipynb 10
def remove_comments(code):
    return "".join(token[2] for token in _lex_tokens(code) if token[1] not in Comment)

# %% ../nbs/04_code_utils.ipynb 18
def _end_block_followed_by_non_comment_block(code):
    return (
        code.rfind('*/') != -1
        and not ends_in_comment_block(code[code.rfind('*/')+2:])
    )

# %% ../nbs/04_code_utils.ipynb 19
def ends_in_comment_block(code):
    last_token = list(_lex_tokens(code))[-1]
    last_token_type = last_token[1]
    return (
        last_token_type == Comment.Multiline
        and code.strip()[-2:] != "*/"
        and not _end_block_followed_by_non_comment_block(code)
    )

# %% ../nbs/04_code_utils.ipynb 23
def _is_not_cr_delimiter(delimiter):
    return delimiter != 'cr'

# %% ../nbs/04_code_utils.ipynb 24
delimit_regex = re.compile(r'^[ \t]*#delimit(.*$)', flags=re.MULTILINE)
def _replace_delimiter(code, sc_delimiter=False):
    # Recursively replace custom delimiter with newline

    split = delimit_regex.split(code.strip(), maxsplit=1)

    if len(split) == 3:
        before = split[0]
        after = _replace_delimiter(split[2], _is_not_cr_delimiter(split[1].strip()))
    else:
        before = code
        after = ''

    if sc_delimiter:
        before_last_sc_pos = before.rfind(';')
        if before_last_sc_pos < len(before.strip()) - 1:
            before = before[:before_last_sc_pos+1]
            if len(split) > 1:
                after = _replace_delimiter(before[before_last_sc_pos+1:]+" ".join(split[1:]), sc_delimiter=True)
        before = before.replace('\r', ' ').replace('\n', ' ')
        before = before.replace(';','\n')

    return before + after

# %% ../nbs/04_code_utils.ipynb 33
def _replace_tabs(code):
    return code.replace("\t", "    ")

# %% ../nbs/04_code_utils.ipynb 35
def valid_single_line_code(code):
    code = _replace_tabs(remove_comments(code))
    if delimit_regex.match(code):
        return ""
    else:
        return code

# %% ../nbs/04_code_utils.ipynb 37
def ending_sc_delimiter(code, sc_delimiter=False):
    code = remove_comments(code)
    # Recursively determine ending delimiter
    split = delimit_regex.split(code.strip(),maxsplit=1)
    
    if len(split) == 3:
        before = split[0]
    else:
        before = code
    if sc_delimiter:
        before_last_sc_pos = before.rfind(';')
        if before_last_sc_pos < len(before.strip()) - 1:
            if len(split) > 1:
                return ending_sc_delimiter(before[before_last_sc_pos+1:]+" ".join(split[1:]), sc_delimiter=True)
            
    if len(split) == 3:
        sc_delimiter = ending_sc_delimiter(split[2], _is_not_cr_delimiter(split[1].strip()))
    elif len(split) == 2:
        sc_delimiter = _is_not_cr_delimiter(split[1].strip())

    return sc_delimiter

# %% ../nbs/04_code_utils.ipynb 44
# Detect Multiple whitespace
multi_regex = re.compile(r'(?P<char>\S) +')

def standardize_code(code, sc_delimiter=False):
    """Remove comments spanning multiple lines and replace custom delimiters"""
    code = remove_comments(code)
    
    # After removing multi-line comments, which could include "#delimit;"
    code = _replace_delimiter(code, sc_delimiter) 
    
    # Replace multiple interior whitespace with one
    code = multi_regex.sub('\g<char> ',code)
    
    # Delete blank lines and whitespace at end of lines
    code_lines = code.splitlines()
    std_lines = []
    for code_line in code_lines:
        cs = code_line.rstrip()
        if cs:
            std_lines.append(cs)
    return '\n'.join(std_lines)

# %% ../nbs/04_code_utils.ipynb 57
def _startswith_stata_abbrev(string, full_command, shortest_abbrev):
    for j in range(len(shortest_abbrev), len(full_command)+1):
        if string.startswith(full_command[0:j] + ' '):
            return True
    return False

# %% ../nbs/04_code_utils.ipynb 59
def _remove_prefixes(std_code_line):
    std_code_line = std_code_line.lstrip()
    if (_startswith_stata_abbrev(std_code_line, 'quietly', 'qui')
        or std_code_line.startswith('capture ')
        or _startswith_stata_abbrev(std_code_line, 'noisily', 'n')):
        return _remove_prefixes(std_code_line.split(None, maxsplit=1)[1])
    else:
        return std_code_line

# %% ../nbs/04_code_utils.ipynb 63
def ending_code_version(code, sc_delimiter=False, code_version=None, stata_version='17.0'):
    if 'version' not in code:
        return code_version
    std_code = standardize_code(code, sc_delimiter)
    for std_code_line in reversed(std_code.splitlines()):
        if 'version ' not in std_code_line:
            continue
        m = re.match(r'\A\s*version ([0-9]+(?:\.[0-9][0-9]?)?)\Z', _remove_prefixes(std_code_line))
        if m:
            _version = Decimal(m.group(1)).normalize()
            if Decimal('1') <= _version <= Decimal(stata_version):
                code_version = None if _version == Decimal(stata_version).normalize() else str(_version)
                break
    return code_version

# %% ../nbs/04_code_utils.ipynb 67
pre = (
    r'(cap(t|tu|tur|ture)?'
    r'|qui(e|et|etl|etly)?'
    r'|n(o|oi|ois|oisi|oisil|oisily)?)')
kwargs = {'flags': re.MULTILINE}
local_def_in = re.compile(
    r"(^\s*({0} )*(loc(a|al)?|tempname|tempvar|tempfile|gettoken|token(i|iz|ize)?|levelsof)\s)|st_local\(".format(pre),
    **kwargs,
).search

# %% ../nbs/04_code_utils.ipynb 69
preserve_restore_in = re.compile(
    r"(^({0} )*(preserve|restore)[,\s]?\.*?$)|(;({0} )*(preserve|restore)[,\s]?\.*?$)".format(pre),
    **kwargs,
).search

# %% ../nbs/04_code_utils.ipynb 72
def is_start_of_program_block(std_code_line):
    cs = _remove_prefixes(std_code_line)
    _starts_program = (_startswith_stata_abbrev(cs, 'program', 'pr')
                       and not (cs.split()[1] in ['di', 'dir', 'drop', 'l', 'li', 'lis', 'list']))
    return (_starts_program
            or (cs in {'mata', 'mata:'})
            or (cs in {'python', 'python:'}))

# %% ../nbs/04_code_utils.ipynb 74
def _prog_blocks(std_code_lines):
    next_block_lines = []
    in_program = False
    for std_code_line in std_code_lines:         
        if is_start_of_program_block(std_code_line):
            if next_block_lines: # previous lines
                yield _block(next_block_lines, is_prog=in_program)
                next_block_lines = []
            in_program = True
        next_block_lines.append(std_code_line)
        if std_code_line == 'end': # regardless of whether in_program
            yield _block(next_block_lines, is_prog=True)
            next_block_lines = []
            in_program = False
    if next_block_lines:
        yield _block(next_block_lines, in_program)
        

def _block(block_lines, is_prog):
    return {"is_prog": is_prog, "std_code": '\n'.join(block_lines)}

# %% ../nbs/04_code_utils.ipynb 75
def break_out_prog_blocks(code, sc_delimiter=False):
    std_code_lines = standardize_code(code, sc_delimiter).splitlines()
    return list(_prog_blocks(std_code_lines))



"""Utilities for loading Stata and nbstata"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_config.ipynb.

# %% auto 0
__all__ = ['find_dir_edition', 'find_edition', 'set_pystata_path', 'launch_stata', 'set_graph_format', 'xdg_user_config_path',
           'old_user_config_path', 'Config']

# %% ../nbs/01_config.ipynb 5
from .misc_utils import print_red
from fastcore.basics import patch_to
import os
import sys
import platform
from shutil import which
from pathlib import Path
from packaging import version
import configparser

# %% ../nbs/01_config.ipynb 8
def _win_find_path(_dir=None):
    if _dir is None:
        dirs = [r'C:\Program Files\Stata19',
                r'C:\Program Files\Stata18',
                r'C:\Program Files\Stata17']
    else:
        dirs = [_dir]
    for this_dir in dirs:
        path = Path(this_dir)
        if os.path.exists(path):
            executables = [exe for exe in path.glob("Stata*.exe") if exe not in set(path.glob("Stata*_old.exe"))]
            if executables:
                return str(executables[0])
    # Otherwise, try old way
    import winreg
    reg = winreg.ConnectRegistry(None, winreg.HKEY_CLASSES_ROOT)
    subkey = r'Stata17Do\shell\do\command'
    try:
        key = winreg.OpenKey(reg, subkey)
        return winreg.QueryValue(key, None).split('"')[1]
    except FileNotFoundError:
        return ''

# %% ../nbs/01_config.ipynb 10
def _mac_find_path(_dir=None):
    """
    Attempt to find Stata path on macOS when not on user's PATH.
    Modified from stata_kernel's original to only location "Applications/Stata". 

    Returns:
        (str): Path to Stata. Empty string if not found.
    """
    if _dir is None:
        _dir = '/Applications/Stata'
    path = Path(_dir)
    if not os.path.exists(path):
        return ''
    else:
        try:
            # find the application with the suffix .app
            # example path: /Applications/Stata/StataMP.app
            return str(next(path.glob("Stata*.app")))
        except StopIteration:
            return ''

# %% ../nbs/01_config.ipynb 12
def _other_find_path():
    for i in ['stata-mp', 'stata-se', 'stata']:
        stata_path = which(i)
        if stata_path:
            return stata_path
    return ''

# %% ../nbs/01_config.ipynb 14
def _find_path(_dir=None):
    if os.getenv('CONTINUOUS_INTEGRATION'):
        print('WARNING: Running as CI; Stata path not set correctly')
        return 'stata'
    path = ''
    if platform.system() == 'Windows':
        path = _win_find_path(_dir)
    elif platform.system() == 'Darwin':
        path = _mac_find_path(_dir)
    return path if path else _other_find_path()

# %% ../nbs/01_config.ipynb 16
def _edition(stata_exe):
    edition = 'be'
    for e in ('be', 'se', 'mp'):
        if stata_exe.find(e) > -1:
            edition = e
            break
    return edition

# %% ../nbs/01_config.ipynb 18
def find_dir_edition(stata_path=None):
    if stata_path is None:
        stata_path = _find_path()
    if not stata_path:
        raise OSError("Stata path not found.")
    stata_dir = str(os.path.dirname(stata_path))
    stata_exe = str(os.path.basename(stata_path)).lower()
    return stata_dir, _edition(stata_exe)

# %% ../nbs/01_config.ipynb 20
def find_edition(stata_dir):
    stata_path = _find_path(stata_dir)
    stata_exe = str(os.path.basename(stata_path)).lower()
    return _edition(stata_exe)

# %% ../nbs/01_config.ipynb 25
def set_pystata_path(stata_dir=None):
    if stata_dir is None:
        stata_dir, _ = find_dir_edition()
    if not os.path.isdir(stata_dir):
        raise OSError(f'Specified stata_dir, "{stata_dir}", is not a valid directory path')
    if not os.path.isdir(os.path.join(stata_dir, 'utilities')):
        raise OSError(f'Specified stata_dir, "{stata_dir}", is not Stata\'s installation path')
    sys.path.append(os.path.join(stata_dir, 'utilities'))

# %% ../nbs/01_config.ipynb 29
def launch_stata(stata_dir=None, edition=None, splash=True):
    """
    We modify stata_setup to make splash screen optional
    """
    if stata_dir is None:
        stata_dir, edition_found = find_dir_edition()
        edition = edition_found if edition is None else edition
    elif edition is None:
        edition = find_edition(stata_dir)
    set_pystata_path(stata_dir)
    import pystata
    try:
        if version.parse(pystata.__version__) >= version.parse("0.1.1"):
            # Splash message control is a new feature of pystata-0.1.1
            pystata.config.init(edition, splash=splash)
        else:
            pystata.config.init(edition)
    except FileNotFoundError as err:
        raise OSError(f'Specified edition, "{edition}", is not present at "{stata_dir}"')

# %% ../nbs/01_config.ipynb 35
def set_graph_format(gformat):
    import pystata
    if gformat == 'pystata':
        gformat = 'svg' # pystata default
    pystata.config.set_graph_format(gformat)

# %% ../nbs/01_config.ipynb 37
def _set_graph_size(width, height):
    import pystata
    pystata.config.set_graph_size(width, height)

# %% ../nbs/01_config.ipynb 41
def _get_config_settings(cpath):
    parser = configparser.ConfigParser(
        empty_lines_in_values=False,
        comment_prefixes=('*','//'),
        inline_comment_prefixes=('//',),
    )
    parser.read(str(cpath))
    return dict(parser.items('nbstata'))

# %% ../nbs/01_config.ipynb 42
def xdg_user_config_path():
    xdg_config_home = Path(os.environ.get('XDG_CONFIG_HOME', Path.home() / '.config'))
    return xdg_config_home / 'nbstata/nbstata.conf'

def old_user_config_path():
    return Path('~/.nbstata.conf').expanduser()

# %% ../nbs/01_config.ipynb 45
class Config:
    "nbstata configuration"
    env = {'stata_dir': None,
           'edition': None,
            'splash': 'False',
           'graph_format': 'png',
           'graph_width': '5.5in',
           'graph_height': '4in',
           'echo': 'None',
           'missing': '.',
          }
    valid_values_of = dict(
        edition={None, 'mp', 'se', 'be'},
        graph_format={'pystata', 'svg', 'png', 'pdf'},
        echo={'True', 'False', 'None'},
        splash={'True', 'False'},
    )
    
    @property
    def splash(self):
        return False if self.env['splash'] == 'False' else True
    
    @property
    def noecho(self):
        return self.env['echo'] == 'None'
    
    @property
    def echo(self):
        return self.env['echo'] == 'True'
    
    def __init__(self):
        """First check if a configuration file exists. If not, try `find_dir_edition`."""
        self.errors = []
        self._update_backup_graph_size()
        self.config_path = None

    def _update_backup_graph_size(self):
        self.backup_graph_size = {key: self.env[key] for key in {'graph_width', 'graph_height'}}
                
    def process_config_file(self):
        global_config_path = Path(os.path.join(sys.prefix, 'etc', 'nbstata.conf'))
        for cpath in (xdg_user_config_path(), old_user_config_path(), global_config_path):      
            if cpath.is_file():
                self._get_config_env(cpath)
                break
            
    def _get_config_env(self, cpath):
        try:
            settings = _get_config_settings(cpath)
        except configparser.Error as err:
            print_red(f"Configuration error in {cpath}:\n"
                      f"    {str(err)}")
        else:
            self.config_path = str(cpath)
            self.update(
                settings, 
                init=True, 
                error_header=f"Configuration errors in {self.config_path}:"  
            )
            
    def update(self, env, init=False, error_header="%set error(s):"):
        init_only_settings = {'stata_dir','edition','splash'}
        allowed_settings = self.env if init else set(self.env)-init_only_settings
        for key in list(env):
            if key not in allowed_settings:
                explanation = (
                    "is only allowed in a configuration file." if key in init_only_settings
                    else "is not a valid setting."
                )
                self.errors.append(f"    '{key}' {explanation}")
                env.pop(key)
            elif key in self.valid_values_of and env[key] not in self.valid_values_of[key]:
                self.errors.append(
                    f"    '{key}' configuration invalid: '{env[key]}' is not a valid value. "
                    f"Reverting to: {key} = {self.env[key]}"
                )
                env.pop(key)
        self._display_and_clear_update_errors(error_header)
        for key in set(env)-{'graph_width', 'graph_height'}:
            if not init: print(f"{key} was {self.env[key]}, is now {env[key]}")
        self.env.update(env)
  
    def _display_and_clear_update_errors(self, error_header):
        if self.errors:
            print_red(error_header)
        for message in self.errors:
            print_red(message)
        self.errors = []
        
    def display_status(self):
        import pystata
        pystata.config.status()
        print(f"""
      echo                   {self.env['echo']}
      missing                {self.env['missing']}
      config file path       {self.config_path}""")

# %% ../nbs/01_config.ipynb 53
@patch_to(Config)
def set_graph_size(self, init=False):
    try:
        _set_graph_size(self.env['graph_width'], self.env['graph_height'])
    except ValueError as err:
        self.env.update(self.backup_graph_size)
        print_red(f"Configuration error: {str(err)}. Graph size not changed.")
        if init: self.set_graph_size() # ensures set to definite measures rather than "default"
    else:
        if {key: self.env[key] for key in {'graph_width', 'graph_height'}} != self.backup_graph_size:
            if not init:
                print(f"graph size was ({self.backup_graph_size['graph_width']}, "
                      f"{self.backup_graph_size['graph_height']}), "
                      f"is now ({self.env['graph_width']}, {self.env['graph_height']}).")
            self._update_backup_graph_size()

# %% ../nbs/01_config.ipynb 59
@patch_to(Config)
def update_graph_config(self, init=False):
    graph_format = self.env['graph_format']
    if graph_format == 'pystata':
        graph_format = 'svg'
    set_graph_format(graph_format)
    self.set_graph_size(init)

# %% ../nbs/01_config.ipynb 61
@patch_to(Config)
def init_stata(self):
    launch_stata(self.env['stata_dir'],
                 self.env['edition'],
                 self.splash,
                )
    self.update_graph_config(init=True)



"""Helper functions that expand on `pystata`/`sfi` functionality"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_stata_more.ipynb.

# %% auto 0
__all__ = ['run_direct_cleaned', 'run_sfi', 'SelectVar', 'IndexVar', 'run_as_program', 'diverted_stata_output',
           'diverted_stata_output_quicker', 'var_from_varlist', 'local_names', 'get_local_dict',
           'locals_code_from_dict', 'user_expression']

# %% ../nbs/03_stata_more.ipynb 4
from .misc_utils import print_red
from .stata import run_direct, run_single, get_local, set_local, drop_var, stata_formatted
from textwrap import dedent
import functools
from contextlib import redirect_stdout
from io import StringIO
import re

# %% ../nbs/03_stata_more.ipynb 8
def run_direct_cleaned(cmds, quietly=False, echo=False, inline=True):
    if quietly:
        with redirect_stdout(StringIO()) as diverted: # to prevent blank line output, as with `program define`
            out = run_direct(cmds, quietly, echo, inline)
            prints = diverted.getvalue()
        for line in prints.splitlines():
            if line.strip():
                print(line)
        return out
    elif len(cmds.splitlines()) > 1:
        with redirect_stdout(StringIO()) as diverted:
            run_direct(cmds, quietly, echo, inline)
            output_lines = diverted.getvalue().splitlines()
        if (len(output_lines) >= 2 
            and not output_lines[0].strip() 
            and "\n".join(output_lines[-2:]).strip() == "."):
            print("\n".join(output_lines[1:-2]))
        else:
            print("\n".join(output_lines))
    else:
        return run_direct(cmds, quietly, echo, inline)

# %% ../nbs/03_stata_more.ipynb 30
def run_sfi(std_code, echo=False, show_exc_warning=True):
    import sfi
    cmds = std_code.splitlines()
    for i, cmd in enumerate(cmds):
        try:
            sfi.SFIToolkit.stata(cmd, echo)
        except Exception as e:
            if show_exc_warning:
                print_red(f"run_sfi (sfi.SFIToolkit.stata) error: {repr(e)}")
            remaining_code = "\n".join(cmds[i:])
            run_direct(remaining_code, echo=echo)
            break

# %% ../nbs/03_stata_more.ipynb 35
class SelectVar():
    """Class for generating Stata select_var for getAsDict"""
    varname = None
    
    def __init__(self, stata_if_code):
        import sfi
        condition = stata_if_code.replace('if ', '', 1).strip()
        if condition:
            self.varname = sfi.SFIToolkit.getTempName()
            cmd = f"quietly gen {self.varname} = cond({condition},1,0)"
            run_single(cmd)

    def clear(self):
        """Remove temporary select_var from Stata dataset"""
        if self.varname:
            drop_var(self.varname)
            
    def __enter__(self):
        return self.varname
    
    def __exit__(self, exc_type, exc_value, exc_tb):
        self.clear()

# %% ../nbs/03_stata_more.ipynb 38
class IndexVar:
    """Class for generating Stata index var for use with pandas"""
    def __enter__(self):
        import sfi
        self.idx_var = sfi.SFIToolkit.getTempName()
        run_single(f"gen {self.idx_var} = _n")
        return self.idx_var
    
    def __exit__(self, exc_type, exc_value, exc_tb):
        drop_var(self.idx_var)

# %% ../nbs/03_stata_more.ipynb 51
def run_as_program(std_non_prog_code, prog_def_option_code=""):
    _program_name = "temp_nbstata_program_name"
    _options = f", {prog_def_option_code}" if prog_def_option_code else ""
    _program_define_code = (
        f"program {_program_name}{_options}\n"
        f"{std_non_prog_code}\n"
        "end\n"
    )
    try:
        run_direct_cleaned(_program_define_code, quietly=True)
        run_direct(_program_name, quietly=False, inline=True, echo=False)
    finally:
        run_single(f"capture program drop {_program_name}")

# %% ../nbs/03_stata_more.ipynb 65
def diverted_stata_output(std_code, runner=None):
    if runner is None:
        runner = functools.partial(run_direct, quietly=False, inline=True, echo=False)
    with redirect_stdout(StringIO()) as diverted:
        run_as_program("return add\ncapture log off", prog_def_option_code="rclass")
        try:
            runner(std_code)
        finally:
            run_as_program("return add\ncapture log on", prog_def_option_code="rclass")
        out = diverted.getvalue()
    return out

# %% ../nbs/03_stata_more.ipynb 71
def diverted_stata_output_quicker(std_non_prog_code):
    with redirect_stdout(StringIO()) as diverted:
        code = f"return add\ncapture log off\n{std_non_prog_code}\ncapture log on"""
        try:
            run_as_program(code, prog_def_option_code="rclass")
        except SystemError as e:
            run_as_program("return add\ncapture log on", prog_def_option_code="rclass")
            raise(e)
        out = diverted.getvalue()
    return out

# %% ../nbs/03_stata_more.ipynb 76
def var_from_varlist(varlist, stfr=None):
    if stfr:
        var_code = varlist.strip()
    else:
        _program_name = "temp_nbstata_varlist_name"
        run_direct_cleaned((
            f"program define {_program_name}\n"
            """ syntax [varlist(default=none)]
                foreach var in `varlist' {
                    disp "`var'"
                }
            end
            """), quietly=True)
        try:
            var_code = diverted_stata_output_quicker(f"""\
                {_program_name} {varlist}
                program drop {_program_name}
                """).strip()
        except Exception as e:
            run_sfi(f"capture program drop {_program_name}")
            raise(e)
    return [c.strip() for c in var_code.split() if c] if var_code else None

# %% ../nbs/03_stata_more.ipynb 86
def local_names():
    run_single("""\
        mata : st_local("temp_nbstata_all_locals", invtokens(st_dir("local", "macro", "*")'))""")
    out = get_local('temp_nbstata_all_locals')
    set_local('temp_nbstata_all_locals', "")
    return out.split()

# %% ../nbs/03_stata_more.ipynb 90
def get_local_dict(_local_names=None):
    if _local_names is None:
        _local_names = local_names()
    return {n: get_local(n) for n in _local_names}

# %% ../nbs/03_stata_more.ipynb 92
def locals_code_from_dict(preexisting_local_dict):
    local_defs = (f"""local {name} `"{preexisting_local_dict[name]}"'"""
                  for name in preexisting_local_dict)
    return "\n".join(local_defs)

# %% ../nbs/03_stata_more.ipynb 98
def user_expression(input_str):
    run_single("tempname output")
    try:
        run_single(f"local `output': display {input_str}")
    except SyntaxError as e:
        combined_message = f"{str(e)}\nInvalid Stata '[%fmt] [=]exp' display expression: {input_str}"
        raise SyntaxError(combined_message)
    return get_local(get_local("output"))



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/nbstata',
                'doc_host': 'https://hugetim.github.io',
                'git_url': 'https://github.com/hugetim/nbstata',
                'lib_path': 'nbstata'},
  'syms': { 'nbstata.browse': { 'nbstata.browse._get_pos_stata_obs_num': ('browse.html#_get_pos_stata_obs_num', 'nbstata/browse.py'),
                                'nbstata.browse._get_start_end_strs': ('browse.html#_get_start_end_strs', 'nbstata/browse.py'),
                                'nbstata.browse._parse_browse_magic_syntax': ( 'browse.html#_parse_browse_magic_syntax',
                                                                               'nbstata/browse.py'),
                                'nbstata.browse._parse_df_params': ('browse.html#_parse_df_params', 'nbstata/browse.py'),
                                'nbstata.browse.browse_df_params': ('browse.html#browse_df_params', 'nbstata/browse.py'),
                                'nbstata.browse.display_df_as_ipydatagrid': ('browse.html#display_df_as_ipydatagrid', 'nbstata/browse.py'),
                                'nbstata.browse.get_df': ('browse.html#get_df', 'nbstata/browse.py'),
                                'nbstata.browse.headtail_df_params': ('browse.html#headtail_df_params', 'nbstata/browse.py'),
                                'nbstata.browse.headtail_get_df': ('browse.html#headtail_get_df', 'nbstata/browse.py'),
                                'nbstata.browse.in_range': ('browse.html#in_range', 'nbstata/browse.py'),
                                'nbstata.browse.parse_browse_magic': ('browse.html#parse_browse_magic', 'nbstata/browse.py'),
                                'nbstata.browse.set_ipydatagrid_height': ('browse.html#set_ipydatagrid_height', 'nbstata/browse.py')},
            'nbstata.cell': { 'nbstata.cell.Cell': ('cell.html#cell', 'nbstata/cell.py'),
                              'nbstata.cell.Cell.__init__': ('cell.html#cell.__init__', 'nbstata/cell.py'),
                              'nbstata.cell.Cell.run': ('cell.html#cell.run', 'nbstata/cell.py')},
            'nbstata.code_utils': { 'nbstata.code_utils._block': ('code_utils.html#_block', 'nbstata/code_utils.py'),
                                    'nbstata.code_utils._end_block_followed_by_non_comment_block': ( 'code_utils.html#_end_block_followed_by_non_comment_block',
                                                                                                     'nbstata/code_utils.py'),
                                    'nbstata.code_utils._is_not_cr_delimiter': ( 'code_utils.html#_is_not_cr_delimiter',
                                                                                 'nbstata/code_utils.py'),
                                    'nbstata.code_utils._lex_tokens': ('code_utils.html#_lex_tokens', 'nbstata/code_utils.py'),
                                    'nbstata.code_utils._prog_blocks': ('code_utils.html#_prog_blocks', 'nbstata/code_utils.py'),
                                    'nbstata.code_utils._remove_prefixes': ('code_utils.html#_remove_prefixes', 'nbstata/code_utils.py'),
                                    'nbstata.code_utils._replace_delimiter': ( 'code_utils.html#_replace_delimiter',
                                                                               'nbstata/code_utils.py'),
                                    'nbstata.code_utils._replace_tabs': ('code_utils.html#_replace_tabs', 'nbstata/code_utils.py'),
                                    'nbstata.code_utils._startswith_stata_abbrev': ( 'code_utils.html#_startswith_stata_abbrev',
                                                                                     'nbstata/code_utils.py'),
                                    'nbstata.code_utils.break_out_prog_blocks': ( 'code_utils.html#break_out_prog_blocks',
                                                                                  'nbstata/code_utils.py'),
                                    'nbstata.code_utils.ending_code_version': ( 'code_utils.html#ending_code_version',
                                                                                'nbstata/code_utils.py'),
                                    'nbstata.code_utils.ending_sc_delimiter': ( 'code_utils.html#ending_sc_delimiter',
                                                                                'nbstata/code_utils.py'),
                                    'nbstata.code_utils.ends_in_comment_block': ( 'code_utils.html#ends_in_comment_block',
                                                                                  'nbstata/code_utils.py'),
                                    'nbstata.code_utils.is_start_of_program_block': ( 'code_utils.html#is_start_of_program_block',
                                                                                      'nbstata/code_utils.py'),
                                    'nbstata.code_utils.remove_comments': ('code_utils.html#remove_comments', 'nbstata/code_utils.py'),
                                    'nbstata.code_utils.standardize_code': ('code_utils.html#standardize_code', 'nbstata/code_utils.py'),
                                    'nbstata.code_utils.valid_single_line_code': ( 'code_utils.html#valid_single_line_code',
                                                                                   'nbstata/code_utils.py')},
            'nbstata.completion_env': { 'nbstata.completion_env.CompletionEnv': ( 'completion_env.html#completionenv',
                                                                                  'nbstata/completion_env.py'),
                                        'nbstata.completion_env.CompletionEnv.__init__': ( 'completion_env.html#completionenv.__init__',
                                                                                           'nbstata/completion_env.py'),
                                        'nbstata.completion_env.CompletionEnv._last_line_first_word': ( 'completion_env.html#completionenv._last_line_first_word',
                                                                                                        'nbstata/completion_env.py'),
                                        'nbstata.completion_env.CompletionEnv._scalar_f_pos_rcomp': ( 'completion_env.html#completionenv._scalar_f_pos_rcomp',
                                                                                                      'nbstata/completion_env.py'),
                                        'nbstata.completion_env.CompletionEnv._start_of_last_chunk': ( 'completion_env.html#completionenv._start_of_last_chunk',
                                                                                                       'nbstata/completion_env.py'),
                                        'nbstata.completion_env.CompletionEnv._start_of_last_word': ( 'completion_env.html#completionenv._start_of_last_word',
                                                                                                      'nbstata/completion_env.py'),
                                        'nbstata.completion_env.CompletionEnv.get_env': ( 'completion_env.html#completionenv.get_env',
                                                                                          'nbstata/completion_env.py'),
                                        'nbstata.completion_env.Env': ('completion_env.html#env', 'nbstata/completion_env.py'),
                                        'nbstata.completion_env._ends_in_a_comment': ( 'completion_env.html#_ends_in_a_comment',
                                                                                       'nbstata/completion_env.py'),
                                        'nbstata.completion_env._ends_in_string_literal': ( 'completion_env.html#_ends_in_string_literal',
                                                                                            'nbstata/completion_env.py'),
                                        'nbstata.completion_env._last_token': ( 'completion_env.html#_last_token',
                                                                                'nbstata/completion_env.py'),
                                        'nbstata.completion_env._last_token_full_string': ( 'completion_env.html#_last_token_full_string',
                                                                                            'nbstata/completion_env.py'),
                                        'nbstata.completion_env._lex_tokens': ( 'completion_env.html#_lex_tokens',
                                                                                'nbstata/completion_env.py')},
            'nbstata.completions': { 'nbstata.completions.CompletionsManager': ( 'completions.html#completionsmanager',
                                                                                 'nbstata/completions.py'),
                                     'nbstata.completions.CompletionsManager.__init__': ( 'completions.html#completionsmanager.__init__',
                                                                                          'nbstata/completions.py'),
                                     'nbstata.completions.CompletionsManager.do': ( 'completions.html#completionsmanager.do',
                                                                                    'nbstata/completions.py'),
                                     'nbstata.completions.CompletionsManager.get': ( 'completions.html#completionsmanager.get',
                                                                                     'nbstata/completions.py'),
                                     'nbstata.completions.CompletionsManager.get_file_paths': ( 'completions.html#completionsmanager.get_file_paths',
                                                                                                'nbstata/completions.py'),
                                     'nbstata.completions.CompletionsManager.get_globals': ( 'completions.html#completionsmanager.get_globals',
                                                                                             'nbstata/completions.py')},
            'nbstata.config': { 'nbstata.config.Config': ('config.html#config', 'nbstata/config.py'),
                                'nbstata.config.Config.__init__': ('config.html#config.__init__', 'nbstata/config.py'),
                                'nbstata.config.Config._display_and_clear_update_errors': ( 'config.html#config._display_and_clear_update_errors',
                                                                                            'nbstata/config.py'),
                                'nbstata.config.Config._get_config_env': ('config.html#config._get_config_env', 'nbstata/config.py'),
                                'nbstata.config.Config._update_backup_graph_size': ( 'config.html#config._update_backup_graph_size',
                                                                                     'nbstata/config.py'),
                                'nbstata.config.Config.display_status': ('config.html#config.display_status', 'nbstata/config.py'),
                                'nbstata.config.Config.echo': ('config.html#config.echo', 'nbstata/config.py'),
                                'nbstata.config.Config.init_stata': ('config.html#config.init_stata', 'nbstata/config.py'),
                                'nbstata.config.Config.noecho': ('config.html#config.noecho', 'nbstata/config.py'),
                                'nbstata.config.Config.process_config_file': ( 'config.html#config.process_config_file',
                                                                               'nbstata/config.py'),
                                'nbstata.config.Config.set_graph_size': ('config.html#config.set_graph_size', 'nbstata/config.py'),
                                'nbstata.config.Config.splash': ('config.html#config.splash', 'nbstata/config.py'),
                                'nbstata.config.Config.update': ('config.html#config.update', 'nbstata/config.py'),
                                'nbstata.config.Config.update_graph_config': ( 'config.html#config.update_graph_config',
                                                                               'nbstata/config.py'),
                                'nbstata.config._edition': ('config.html#_edition', 'nbstata/config.py'),
                                'nbstata.config._find_path': ('config.html#_find_path', 'nbstata/config.py'),
                                'nbstata.config._get_config_settings': ('config.html#_get_config_settings', 'nbstata/config.py'),
                                'nbstata.config._mac_find_path': ('config.html#_mac_find_path', 'nbstata/config.py'),
                                'nbstata.config._other_find_path': ('config.html#_other_find_path', 'nbstata/config.py'),
                                'nbstata.config._set_graph_size': ('config.html#_set_graph_size', 'nbstata/config.py'),
                                'nbstata.config._win_find_path': ('config.html#_win_find_path', 'nbstata/config.py'),
                                'nbstata.config.find_dir_edition': ('config.html#find_dir_edition', 'nbstata/config.py'),
                                'nbstata.config.find_edition': ('config.html#find_edition', 'nbstata/config.py'),
                                'nbstata.config.launch_stata': ('config.html#launch_stata', 'nbstata/config.py'),
                                'nbstata.config.old_user_config_path': ('config.html#old_user_config_path', 'nbstata/config.py'),
                                'nbstata.config.set_graph_format': ('config.html#set_graph_format', 'nbstata/config.py'),
                                'nbstata.config.set_pystata_path': ('config.html#set_pystata_path', 'nbstata/config.py'),
                                'nbstata.config.xdg_user_config_path': ('config.html#xdg_user_config_path', 'nbstata/config.py')},
            'nbstata.inspect': {'nbstata.inspect.get_inspect': ('inspect.html#get_inspect', 'nbstata/inspect.py')},
            'nbstata.install': { 'nbstata.install._conf_default': ('install.html#_conf_default', 'nbstata/install.py'),
                                 'nbstata.install._conf_path': ('install.html#_conf_path', 'nbstata/install.py'),
                                 'nbstata.install._find_stata': ('install.html#_find_stata', 'nbstata/install.py'),
                                 'nbstata.install._is_root': ('install.html#_is_root', 'nbstata/install.py'),
                                 'nbstata.install.create_conf_if_needed': ('install.html#create_conf_if_needed', 'nbstata/install.py'),
                                 'nbstata.install.install_kernel_spec': ('install.html#install_kernel_spec', 'nbstata/install.py'),
                                 'nbstata.install.main': ('install.html#main', 'nbstata/install.py')},
            'nbstata.kernel': { 'nbstata.kernel.PyStataKernel': ('kernel.html#pystatakernel', 'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.__init__': ('kernel.html#pystatakernel.__init__', 'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.do_complete': ('kernel.html#pystatakernel.do_complete', 'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.do_execute': ('kernel.html#pystatakernel.do_execute', 'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.do_history': ('kernel.html#pystatakernel.do_history', 'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.do_inspect': ('kernel.html#pystatakernel.do_inspect', 'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.do_is_complete': ( 'kernel.html#pystatakernel.do_is_complete',
                                                                                 'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.init_session': ( 'kernel.html#pystatakernel.init_session',
                                                                               'nbstata/kernel.py'),
                                'nbstata.kernel.PyStataKernel.post_do_hook': ( 'kernel.html#pystatakernel.post_do_hook',
                                                                               'nbstata/kernel.py'),
                                'nbstata.kernel._format_user_obj': ('kernel.html#_format_user_obj', 'nbstata/kernel.py'),
                                'nbstata.kernel._handle_stata_error': ('kernel.html#_handle_stata_error', 'nbstata/kernel.py'),
                                'nbstata.kernel._handle_stata_import_error': ( 'kernel.html#_handle_stata_import_error',
                                                                               'nbstata/kernel.py'),
                                'nbstata.kernel._handle_stata_init_error': ('kernel.html#_handle_stata_init_error', 'nbstata/kernel.py'),
                                'nbstata.kernel._stata_error_reply': ('kernel.html#_stata_error_reply', 'nbstata/kernel.py'),
                                'nbstata.kernel._user_expressions': ('kernel.html#_user_expressions', 'nbstata/kernel.py'),
                                'nbstata.kernel.print_stata_error': ('kernel.html#print_stata_error', 'nbstata/kernel.py')},
            'nbstata.magics': { 'nbstata.magics.Frame': ('magics.html#frame', 'nbstata/magics.py'),
                                'nbstata.magics.Frame.__enter__': ('magics.html#frame.__enter__', 'nbstata/magics.py'),
                                'nbstata.magics.Frame.__exit__': ('magics.html#frame.__exit__', 'nbstata/magics.py'),
                                'nbstata.magics.Frame.__init__': ('magics.html#frame.__init__', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics': ('magics.html#statamagics', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics._do_magic': ('magics.html#statamagics._do_magic', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics._get_help_html': ( 'magics.html#statamagics._get_help_html',
                                                                               'nbstata/magics.py'),
                                'nbstata.magics.StataMagics._headtail_html': ( 'magics.html#statamagics._headtail_html',
                                                                               'nbstata/magics.py'),
                                'nbstata.magics.StataMagics._magic_frheadtail': ( 'magics.html#statamagics._magic_frheadtail',
                                                                                  'nbstata/magics.py'),
                                'nbstata.magics.StataMagics._magic_headtail': ( 'magics.html#statamagics._magic_headtail',
                                                                                'nbstata/magics.py'),
                                'nbstata.magics.StataMagics._parse_code_for_magic': ( 'magics.html#statamagics._parse_code_for_magic',
                                                                                      'nbstata/magics.py'),
                                'nbstata.magics.StataMagics._unabbrev_magic_name': ( 'magics.html#statamagics._unabbrev_magic_name',
                                                                                     'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic': ('magics.html#statamagics.magic', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_browse': ('magics.html#statamagics.magic_browse', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_delimit': ('magics.html#statamagics.magic_delimit', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_echo': ('magics.html#statamagics.magic_echo', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_frbrowse': ( 'magics.html#statamagics.magic_frbrowse',
                                                                               'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_frhead': ('magics.html#statamagics.magic_frhead', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_frtail': ('magics.html#statamagics.magic_frtail', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_head': ('magics.html#statamagics.magic_head', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_help': ('magics.html#statamagics.magic_help', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_locals': ('magics.html#statamagics.magic_locals', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_noecho': ('magics.html#statamagics.magic_noecho', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_quietly': ('magics.html#statamagics.magic_quietly', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_set': ('magics.html#statamagics.magic_set', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_status': ('magics.html#statamagics.magic_status', 'nbstata/magics.py'),
                                'nbstata.magics.StataMagics.magic_tail': ('magics.html#statamagics.magic_tail', 'nbstata/magics.py'),
                                'nbstata.magics._clean_error_message': ('magics.html#_clean_error_message', 'nbstata/magics.py'),
                                'nbstata.magics._construct_abbrev_dict': ('magics.html#_construct_abbrev_dict', 'nbstata/magics.py'),
                                'nbstata.magics._formatted_local_list': ('magics.html#_formatted_local_list', 'nbstata/magics.py'),
                                'nbstata.magics._get_html_data': ('magics.html#_get_html_data', 'nbstata/magics.py'),
                                'nbstata.magics._get_new_settings': ('magics.html#_get_new_settings', 'nbstata/magics.py'),
                                'nbstata.magics._parse_frame_prefix': ('magics.html#_parse_frame_prefix', 'nbstata/magics.py'),
                                'nbstata.magics._parse_magic_name_code': ('magics.html#_parse_magic_name_code', 'nbstata/magics.py'),
                                'nbstata.magics._process_new_settings': ('magics.html#_process_new_settings', 'nbstata/magics.py'),
                                'nbstata.magics.print_kernel': ('magics.html#print_kernel', 'nbstata/magics.py')},
            'nbstata.misc_utils': { 'nbstata.misc_utils.Timer': ('misc_utils.html#timer', 'nbstata/misc_utils.py'),
                                    'nbstata.misc_utils.Timer.__enter__': ('misc_utils.html#timer.__enter__', 'nbstata/misc_utils.py'),
                                    'nbstata.misc_utils.Timer.__exit__': ('misc_utils.html#timer.__exit__', 'nbstata/misc_utils.py'),
                                    'nbstata.misc_utils.Timer.start': ('misc_utils.html#timer.start', 'nbstata/misc_utils.py'),
                                    'nbstata.misc_utils.Timer.stop': ('misc_utils.html#timer.stop', 'nbstata/misc_utils.py'),
                                    'nbstata.misc_utils.print_red': ('misc_utils.html#print_red', 'nbstata/misc_utils.py')},
            'nbstata.noecho': { 'nbstata.noecho._after_local_dict': ('noecho.html#_after_local_dict', 'nbstata/noecho.py'),
                                'nbstata.noecho._local_names_from_sreturn': ('noecho.html#_local_names_from_sreturn', 'nbstata/noecho.py'),
                                'nbstata.noecho._restore_locals_and_clear_sreturn': ( 'noecho.html#_restore_locals_and_clear_sreturn',
                                                                                      'nbstata/noecho.py'),
                                'nbstata.noecho._run_as_program_w_locals_sreturned': ( 'noecho.html#_run_as_program_w_locals_sreturned',
                                                                                       'nbstata/noecho.py'),
                                'nbstata.noecho.run_as_program_w_locals': ('noecho.html#run_as_program_w_locals', 'nbstata/noecho.py'),
                                'nbstata.noecho.run_noecho': ('noecho.html#run_noecho', 'nbstata/noecho.py'),
                                'nbstata.noecho.run_non_prog_noecho': ('noecho.html#run_non_prog_noecho', 'nbstata/noecho.py'),
                                'nbstata.noecho.run_prog_noecho': ('noecho.html#run_prog_noecho', 'nbstata/noecho.py')},
            'nbstata.pandas': { 'nbstata.pandas._better_dataframe': ('pandas.html#_better_dataframe', 'nbstata/pandas.py'),
                                'nbstata.pandas._simple_dataframe_from_stata': ( 'pandas.html#_simple_dataframe_from_stata',
                                                                                 'nbstata/pandas.py'),
                                'nbstata.pandas.better_dataframe_from_stata': ( 'pandas.html#better_dataframe_from_stata',
                                                                                'nbstata/pandas.py'),
                                'nbstata.pandas.better_pdataframe_from_data': ( 'pandas.html#better_pdataframe_from_data',
                                                                                'nbstata/pandas.py'),
                                'nbstata.pandas.better_pdataframe_from_frame': ( 'pandas.html#better_pdataframe_from_frame',
                                                                                 'nbstata/pandas.py')},
            'nbstata.stata': { 'nbstata.stata.drop_var': ('stata.html#drop_var', 'nbstata/stata.py'),
                               'nbstata.stata.get_global': ('stata.html#get_global', 'nbstata/stata.py'),
                               'nbstata.stata.get_local': ('stata.html#get_local', 'nbstata/stata.py'),
                               'nbstata.stata.get_scalar': ('stata.html#get_scalar', 'nbstata/stata.py'),
                               'nbstata.stata.macro_expand': ('stata.html#macro_expand', 'nbstata/stata.py'),
                               'nbstata.stata.obs_count': ('stata.html#obs_count', 'nbstata/stata.py'),
                               'nbstata.stata.pwd': ('stata.html#pwd', 'nbstata/stata.py'),
                               'nbstata.stata.run_direct': ('stata.html#run_direct', 'nbstata/stata.py'),
                               'nbstata.stata.run_single': ('stata.html#run_single', 'nbstata/stata.py'),
                               'nbstata.stata.set_local': ('stata.html#set_local', 'nbstata/stata.py'),
                               'nbstata.stata.stata_formatted': ('stata.html#stata_formatted', 'nbstata/stata.py'),
                               'nbstata.stata.variable_names': ('stata.html#variable_names', 'nbstata/stata.py')},
            'nbstata.stata_more': { 'nbstata.stata_more.IndexVar': ('stata_more.html#indexvar', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.IndexVar.__enter__': ( 'stata_more.html#indexvar.__enter__',
                                                                               'nbstata/stata_more.py'),
                                    'nbstata.stata_more.IndexVar.__exit__': ('stata_more.html#indexvar.__exit__', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.SelectVar': ('stata_more.html#selectvar', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.SelectVar.__enter__': ( 'stata_more.html#selectvar.__enter__',
                                                                                'nbstata/stata_more.py'),
                                    'nbstata.stata_more.SelectVar.__exit__': ( 'stata_more.html#selectvar.__exit__',
                                                                               'nbstata/stata_more.py'),
                                    'nbstata.stata_more.SelectVar.__init__': ( 'stata_more.html#selectvar.__init__',
                                                                               'nbstata/stata_more.py'),
                                    'nbstata.stata_more.SelectVar.clear': ('stata_more.html#selectvar.clear', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.diverted_stata_output': ( 'stata_more.html#diverted_stata_output',
                                                                                  'nbstata/stata_more.py'),
                                    'nbstata.stata_more.diverted_stata_output_quicker': ( 'stata_more.html#diverted_stata_output_quicker',
                                                                                          'nbstata/stata_more.py'),
                                    'nbstata.stata_more.get_local_dict': ('stata_more.html#get_local_dict', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.local_names': ('stata_more.html#local_names', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.locals_code_from_dict': ( 'stata_more.html#locals_code_from_dict',
                                                                                  'nbstata/stata_more.py'),
                                    'nbstata.stata_more.run_as_program': ('stata_more.html#run_as_program', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.run_direct_cleaned': ( 'stata_more.html#run_direct_cleaned',
                                                                               'nbstata/stata_more.py'),
                                    'nbstata.stata_more.run_sfi': ('stata_more.html#run_sfi', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.user_expression': ('stata_more.html#user_expression', 'nbstata/stata_more.py'),
                                    'nbstata.stata_more.var_from_varlist': ('stata_more.html#var_from_varlist', 'nbstata/stata_more.py')},
            'nbstata.stata_session': { 'nbstata.stata_session.StataSession': ( 'stata_session.html#statasession',
                                                                               'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession.__init__': ( 'stata_session.html#statasession.__init__',
                                                                                        'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession._compile_re': ( 'stata_session.html#statasession._compile_re',
                                                                                           'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession._completions': ( 'stata_session.html#statasession._completions',
                                                                                            'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession._get_locals': ( 'stata_session.html#statasession._get_locals',
                                                                                           'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession._post_run_hook': ( 'stata_session.html#statasession._post_run_hook',
                                                                                              'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession._run_as_program_w_locals': ( 'stata_session.html#statasession._run_as_program_w_locals',
                                                                                                        'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession._update_ending_delimiter': ( 'stata_session.html#statasession._update_ending_delimiter',
                                                                                                        'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession.clear_suggestions': ( 'stata_session.html#statasession.clear_suggestions',
                                                                                                 'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession.dispatch_run': ( 'stata_session.html#statasession.dispatch_run',
                                                                                            'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession.get_local_dict': ( 'stata_session.html#statasession.get_local_dict',
                                                                                              'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession.get_suggestions': ( 'stata_session.html#statasession.get_suggestions',
                                                                                               'nbstata/stata_session.py'),
                                       'nbstata.stata_session.StataSession.refresh_suggestions': ( 'stata_session.html#statasession.refresh_suggestions',
                                                                                                   'nbstata/stata_session.py'),
                                       'nbstata.stata_session._run_simple': ('stata_session.html#_run_simple', 'nbstata/stata_session.py'),
                                       'nbstata.stata_session.warn_re_unclosed_comment_block_if_needed': ( 'stata_session.html#warn_re_unclosed_comment_block_if_needed',
                                                                                                           'nbstata/stata_session.py')}}}



"""General helper functions with no Jupyter or pystata dependence"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_misc_utils.ipynb.

# %% auto 0
__all__ = ['Timer', 'print_red']

# %% ../nbs/00_misc_utils.ipynb 3
import time

# %% ../nbs/00_misc_utils.ipynb 4
class Timer():
    text = "Elapsed time: {:0.4f} seconds"
    logger = print
    _start_time = None

    def start(self):
        self._start_time = time.perf_counter()

    def stop(self):
        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None
        if self.logger:
            self.logger(self.text.format(elapsed_time))

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc_info):
        self.stop()

# %% ../nbs/00_misc_utils.ipynb 6
def print_red(text):
    print(f"\x1b[31m{text}\x1b[0m")



"""Class representing a single code cell"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/13_cell.ipynb.

# %% auto 0
__all__ = ['magic_handler', 'Cell']

# %% ../nbs/13_cell.ipynb 3
from .stata_session import StataSession
from .magics import StataMagics
from fastcore.basics import patch_to

# %% ../nbs/13_cell.ipynb 4
magic_handler = StataMagics()

# %% ../nbs/13_cell.ipynb 5
class Cell:
    """A class for managing execution of a single code cell"""                  
    def __init__(self, kernel, code_w_magics, silent=False):
        self.noecho = kernel.nbstata_config.noecho
        self.echo = kernel.nbstata_config.echo
        self.quietly = silent
        self.stata_session = kernel.stata_session
        self.code = magic_handler.magic(code_w_magics, kernel, self)
       
    def run(self):
        if not self.code:
            return
        self.stata_session.dispatch_run(self.code, 
            quietly=self.quietly, echo=self.echo, noecho=self.noecho)



"""Provides output for kernel.do_inspect()"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/12_inspect.ipynb.

# %% auto 0
__all__ = ['get_inspect']

# %% ../nbs/12_inspect.ipynb 3
from .stata_more import run_as_program, run_sfi, diverted_stata_output
import functools

# %% ../nbs/12_inspect.ipynb 6
def get_inspect(code="", cursor_pos=0, detail_level=0, omit_sections=()):
    runner = functools.partial(run_as_program, prog_def_option_code="rclass")
    inspect_code = """
        disp _newline "*** Stored results:"
        return list
        ereturn list
        return add
        display "*** Last updated `c(current_time)' `c(current_date)' ***"
        describe, fullnames
        """
    raw_output = diverted_stata_output(inspect_code, runner=runner)
    desc_start = raw_output.find('*** Last updated ')
    out = raw_output[desc_start:]
    if desc_start > 21:
        out += raw_output[:desc_start]
    return out



"""Stata-to-pandas utilities, used in `nbstata.browse`"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_pandas.ipynb.

# %% auto 0
__all__ = ['better_dataframe_from_stata', 'better_pdataframe_from_data', 'better_pdataframe_from_frame']

# %% ../nbs/06_pandas.ipynb 3
from .stata import stata_formatted
from .stata_more import IndexVar

# %% ../nbs/06_pandas.ipynb 6
def _better_dataframe(hdl, var, obs, selectvar, valuelabel, missingval):
    import pandas as pd
    with IndexVar() as idx_var:
        data = hdl.getAsDict(var, obs, selectvar, valuelabel, missingval)
        if not data:
            return pd.DataFrame()
    
        if idx_var in data:
            idx = data.pop(idx_var)
        else:
            temp_var = [idx_var, selectvar] if selectvar else idx_var
            idx = hdl.getAsDict(temp_var, obs, selectvar, valuelabel, missingval).pop(idx_var)
        idx = pd.array(idx, dtype='int64')

        return pd.DataFrame(data=data, index=idx)

# %% ../nbs/06_pandas.ipynb 17
def _simple_dataframe_from_stata(stfr, var, valuelabel, missingval):
    from pystata import stata
    if stfr is None:
        df = stata.pdataframe_from_data(var=var, valuelabel=valuelabel, missingval=missingval)
    else:
        df = stata.pdataframe_from_frame(stfr, var=var, valuelabel=valuelabel, missingval=missingval)
    df.index += 1
    return df

# %% ../nbs/06_pandas.ipynb 20
def better_dataframe_from_stata(stfr, var, obs, selectvar, valuelabel, missingval, sformat):
    import numpy as np
    import pandas as pd
    import sfi
    hdl = sfi.Data if stfr is None else sfi.Frame.connect(stfr)
    custom_index_not_needed = obs is None and not selectvar
    if custom_index_not_needed:
        df = _simple_dataframe_from_stata(stfr, var, valuelabel, missingval)
    else:
        if hdl.getObsTotal() <= 0:
            return pd.DataFrame()
        df = _better_dataframe(hdl, var, obs, selectvar, valuelabel, missingval)
    if sformat:
        for v in list(df.columns):
            if hdl.isVarTypeString(v) or (valuelabel and missingval==np.NaN
                                          and not pd.api.types.is_numeric_dtype(df[v])):
                continue
            v_format = hdl.getVarFormat(v)
            if missingval != np.NaN and not pd.api.types.is_numeric_dtype(df[v]):
                def format_value(x):
                    return stata_formatted(x, v_format).lstrip() if type(x)!=str else x
            else:
                def format_value(x):
                    return stata_formatted(x, v_format).lstrip()
            df[v] = df[v].apply(format_value)
    return df

# %% ../nbs/06_pandas.ipynb 21
def better_pdataframe_from_data(var=None, obs=None, selectvar=None, valuelabel=False, missingval=None, sformat=False):
    import numpy as np
    if missingval is None:
        missingval = np.NaN
    return better_dataframe_from_stata(None, var, obs, selectvar, valuelabel, missingval, sformat)

# %% ../nbs/06_pandas.ipynb 22
def better_pdataframe_from_frame(stfr, var=None, obs=None, selectvar=None, valuelabel=False, missingval=None, sformat=False):
    import numpy as np
    if missingval is None:
        missingval = np.NaN
    return better_dataframe_from_stata(stfr, var, obs, selectvar, valuelabel, missingval, sformat)



"""For running multi-line Stata code without echoing the commands"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_noecho.ipynb.

# %% auto 0
__all__ = ['parse_sreturn', 'run_as_program_w_locals', 'run_non_prog_noecho', 'run_prog_noecho', 'run_noecho']

# %% ../nbs/05_noecho.ipynb 4
from .code_utils import break_out_prog_blocks, valid_single_line_code, local_def_in, preserve_restore_in
from .stata import run_direct, set_local, run_single, get_global
from . import stata_more as sm 
from textwrap import dedent
import re

# %% ../nbs/05_noecho.ipynb 7
def _run_as_program_w_locals_sreturned(std_code):
    sreturn_code = dedent("""\
        
        mata : st_local("temp_nbstata_all_locals", invtokens(st_dir("local", "macro", "*")'))
        foreach lname in `temp_nbstata_all_locals' {
            sreturn local `lname' "``lname''"
        }
        """)
    store_new_locals_code = ("sreturn clear\n" 
                             + std_code
                             + sreturn_code)                          
    sm.run_as_program(store_new_locals_code, "sclass")

# %% ../nbs/05_noecho.ipynb 11
parse_sreturn = re.compile(
    r'^\s*?(?:\ss\((?P<name>\w+)\) : )', flags=re.MULTILINE
).findall

# %% ../nbs/05_noecho.ipynb 13
def _local_names_from_sreturn(sreturn_output):
    matches = parse_sreturn(sreturn_output)
    return matches

# %% ../nbs/05_noecho.ipynb 15
def _after_local_dict():
    sreturn_output = sm.diverted_stata_output_quicker("sreturn list")
    _local_names = _local_names_from_sreturn(sreturn_output)
    return {name: get_global(f"s({name})") for name in _local_names}

# %% ../nbs/05_noecho.ipynb 20
def _restore_locals_and_clear_sreturn():
    for lname, value in _after_local_dict().items():
        set_local(lname, value)
    run_single("sreturn clear")

# %% ../nbs/05_noecho.ipynb 22
def run_as_program_w_locals(std_code, local_dict=None):
    if local_dict is None:
        local_dict = sm.get_local_dict()
    locals_code = sm.locals_code_from_dict(local_dict)
    if not local_def_in(std_code):
        sm.run_as_program(f"""{locals_code}\n{std_code}""")
    else:
        _run_as_program_w_locals_sreturned(f"""{locals_code}\n{std_code}""")
        _restore_locals_and_clear_sreturn()

# %% ../nbs/05_noecho.ipynb 25
def run_non_prog_noecho(std_non_prog_code, run_as_prog=run_as_program_w_locals):
    if len(std_non_prog_code.splitlines()) <= 1:  # to keep it simple when we can
        run_direct(valid_single_line_code(std_non_prog_code),
                   quietly=False, inline=True, echo=False)
    elif preserve_restore_in(std_non_prog_code):
        print("(Note: Below code run with echo to enable preserve/restore functionality.)")
        run_direct(std_non_prog_code,
                   quietly=False, inline=True, echo=False)
    else:
        run_as_prog(std_non_prog_code)

# %% ../nbs/05_noecho.ipynb 32
def run_prog_noecho(std_prog_code):
    if std_prog_code.splitlines()[0] in {'mata', 'mata:'}:  # b/c 'quietly' blocks mata output
        run_direct(std_prog_code, quietly=False, inline=True, echo=False)
    else:
        run_direct(std_prog_code, quietly=True, inline=True, echo=False)

# %% ../nbs/05_noecho.ipynb 38
def run_noecho(code, sc_delimiter=False, run_as_prog=run_as_program_w_locals):
    """After `break_out_prog_blocks`, run each prog and non-prog block noecho"""
    for block in break_out_prog_blocks(code, sc_delimiter):
        if block['is_prog']:
            run_prog_noecho(block['std_code'])
        else:
            run_non_prog_noecho(block['std_code'], run_as_prog=run_as_prog)



"""A class for representing a Stata session"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_stata_session.ipynb.

# %% auto 0
__all__ = ['StataSession', 'warn_re_unclosed_comment_block_if_needed']

# %% ../nbs/08_stata_session.ipynb 4
from .misc_utils import print_red
from .stata import run_direct, get_local, get_scalar
from .stata_more import diverted_stata_output_quicker, local_names, run_sfi
from .stata_more import get_local_dict as _get_local_dict
from nbstata.code_utils import (
    valid_single_line_code,
    ending_sc_delimiter,
    ends_in_comment_block,
    ending_code_version,
)
from .noecho import run_as_program_w_locals, run_noecho
from fastcore.basics import patch_to
from textwrap import dedent
import re

# %% ../nbs/08_stata_session.ipynb 6
class StataSession():
    def __init__(self):
        """"""
        self.sc_delimiter = False
        self.code_version = None
        self.stata_version = None
        self.clear_suggestions()
        self._compile_re()

    def clear_suggestions(self):
        self.suggestions = None
        
    def _compile_re(self):
        self.matchall = re.compile(
            r"\A.*?"
            r"^%varlist%(?P<varlist>.*?)"
            r"%globals%(?P<globals>.*?)"
            #r"%locals%(?P<locals>.*?)"
            r"%scalars%(?P<scalars>.*?)"
            r"%matrices%(?P<matrices>.*?)%end%", #"(\Z|---+\s*end)",
            flags=re.DOTALL + re.MULTILINE).match

        # Varlist-style matching; applies to most
        self.varlist = re.compile(r"(?:\s+)(\S+)", flags=re.MULTILINE)

        # file-style matching
        self.filelist = re.compile(r"[\r\n]{1,2}", flags=re.MULTILINE)

        # Clean line-breaks.
        self.varclean = re.compile(
            r"(?=\s*)[\r\n]{1,2}?^>\s", flags=re.MULTILINE).sub
        
        #         # Match output from mata mata desc
#         self.matadesc = re.compile(
#             r"(\A.*?---+|---+[\r\n]*\Z)", flags=re.MULTILINE + re.DOTALL)

#         self.matalist = re.compile(
#             r"(?:.*?)\s(\S+)\s*$", flags=re.MULTILINE + re.DOTALL)

#         self.mataclean = re.compile(r"\W.*?(\b|$)")
#         self.matasearch = re.compile(r"(?P<kw>\w.*?(?=\W|\b|$))").search

# %% ../nbs/08_stata_session.ipynb 7
@patch_to(StataSession)
def refresh_suggestions(self):
    self.suggestions = self.get_suggestions()

# %% ../nbs/08_stata_session.ipynb 8
@patch_to(StataSession)
def _completions(self):
    return diverted_stata_output_quicker(dedent("""\
        local _temp_completions_while_local_ = 1
        while `_temp_completions_while_local_' {
        set more off
        set trace off
        if `"`varlist'"' != "" {
        local _temp_completions_varlist_loc_ `"`varlist'"'
        }
        syntax [varlist]
        disp "%varlist%"
        disp `"`varlist'"'
        macro drop _varlist __temp_completions_while_local_
        if `"`_temp_completions_varlist_loc_'"' != "" {
        local varlist `"`_temp_completions_varlist_loc_'"'
        macro drop __temp_completions_varlist_loc_
        }
        disp "%globals%"
        disp `"`:all globals'"'
        *disp "%locals%"
        *mata : invtokens(st_dir("local", "macro", "*")')
        disp "%scalars%"
        disp `"`:all scalars'"'
        disp "%matrices%"
        disp `"`:all matrices'"'
        disp "%end%"
        local _temp_completions_while_local_ = 0
        }
        macro drop _temp_completions_while_local_
    """))

# %% ../nbs/08_stata_session.ipynb 11
@patch_to(StataSession)
def _get_locals(self):
    return self.suggestions['locals'] if self.suggestions else local_names()

# %% ../nbs/08_stata_session.ipynb 15
@patch_to(StataSession)
def get_suggestions(self):
    match = self.matchall(self._completions())
    suggestions = match.groupdict()
#         suggestions['mata'] = self._parse_mata_desc(suggestions['mata'])
#         suggestions['programs'] = self._parse_programs_desc(
#             suggestions['programs'])
    for k, v in suggestions.items():
        suggestions[k] = self.varlist.findall(self.varclean('', v))
    suggestions['locals'] = self._get_locals()
    return suggestions

# %% ../nbs/08_stata_session.ipynb 19
@patch_to(StataSession)
def get_local_dict(self):
    return _get_local_dict(self._get_locals())

# %% ../nbs/08_stata_session.ipynb 21
@patch_to(StataSession)
def _run_as_program_w_locals(self, std_code):
    """After `break_out_prog_blocks`, run noecho, inserting locals when needed"""
    return run_as_program_w_locals(std_code, local_dict=self.get_local_dict())

# %% ../nbs/08_stata_session.ipynb 26
def _run_simple(code, quietly=False, echo=False, sc_delimiter=False):
    if sc_delimiter:
        code = "#delimit;\n" + code
    if len(code.splitlines()) == 1:
        code = valid_single_line_code(code)
    run_direct(code, quietly=quietly, inline=not quietly, echo=echo)

# %% ../nbs/08_stata_session.ipynb 29
_final_delimiter_warning = (
    "Warning: Code cell (with #delimit; in effect) does not end in ';'. "
    "Exported .do script may behave differently from notebook. "
    "In v1.0, nbstata may trigger an error instead of just a warning."
)

# %% ../nbs/08_stata_session.ipynb 30
@patch_to(StataSession)    
def _update_ending_delimiter(self, code):
    self.sc_delimiter = ending_sc_delimiter(code, self.sc_delimiter)
    _final_character = code.strip()[-1]
    _code_missing_final_delimiter = (self.sc_delimiter
                                     and _final_character != ';')
    if _code_missing_final_delimiter:
        print_red(_final_delimiter_warning)

# %% ../nbs/08_stata_session.ipynb 32
def warn_re_unclosed_comment_block_if_needed(code):
    if ends_in_comment_block(code):
        print_red("Warning: Code cell ends in a comment block without a "
                  "closing '*/'. Exported .do script may behave differently "
                  "from notebook. In v1.0, nbstata may trigger an error "
                  "instead of just a warning."
                 )

# %% ../nbs/08_stata_session.ipynb 35
@patch_to(StataSession)
def _post_run_hook(self, code):
    self.clear_suggestions()
    if self.stata_version is None:
        self.stata_version = f"{get_scalar('c(stata_version)'):0.2f}"
    self.code_version = ending_code_version(code, self.sc_delimiter, self.code_version, self.stata_version)
    self._update_ending_delimiter(code) # after updating code_version (based on starting sc_delimiter)
    warn_re_unclosed_comment_block_if_needed(code)

# %% ../nbs/08_stata_session.ipynb 36
@patch_to(StataSession)
def dispatch_run(self, code, quietly=False, echo=False, noecho=False):
    if self.code_version:
        version_prefix = "capture version " + self.code_version + (";" if self.sc_delimiter else "\n")
        code = version_prefix + code
    if noecho and not quietly:
        run_noecho(code, self.sc_delimiter, run_as_prog=self._run_as_program_w_locals)
    else:
        _run_simple(code, quietly, echo, self.sc_delimiter)
    self._post_run_hook(code)



"""Helpers for browse, head, and tail magics"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_browse.ipynb.

# %% auto 0
__all__ = ['matchparts', 'in_range', 'parse_browse_magic', 'get_df', 'headtail_df_params', 'headtail_get_df', 'browse_df_params',
           'set_ipydatagrid_height', 'display_df_as_ipydatagrid']

# %% ../nbs/07_browse.ipynb 3
from .misc_utils import print_red
from .stata import run_single
from .stata_more import SelectVar, run_direct_cleaned, diverted_stata_output_quicker, run_sfi
from .pandas import better_pdataframe_from_data
from fastcore.basics import patch_to
import re

# %% ../nbs/07_browse.ipynb 6
def _get_start_end_strs(stata_in_code):
    stata_range_code = stata_in_code.replace('in ','').strip()
    slash_pos = stata_range_code.find('/')
    if slash_pos != -1:
        start_str = stata_range_code[:slash_pos]
        end_str = stata_range_code[slash_pos+1:]
    else:
        start_str = "1"
        end_str = stata_range_code
    return start_str, end_str

# %% ../nbs/07_browse.ipynb 8
def _get_pos_stata_obs_num(in_obs_str, count):
    temp_str = in_obs_str.strip().upper()
    if temp_str == 'F': 
        in_obs = 1
    elif temp_str == 'L':
        in_obs = count
    else:
        try:
            in_obs = int(in_obs_str)
        except ValueError as e:
            raise ValueError(f"{in_obs_str} invalid observation number")
        if in_obs < 0: in_obs += count + 1
        if in_obs < 1 or in_obs > count:
            raise ValueError(f"{in_obs_str} invalid observation number")
    return in_obs

# %% ../nbs/07_browse.ipynb 10
def in_range(stata_in_code, count):
    """Return in-statement range"""
    if not stata_in_code.strip():
        return (None, None)
    start, end = (_get_pos_stata_obs_num(in_str, count)
                  for in_str in _get_start_end_strs(stata_in_code))
    if start > end:
        raise ValueError("observations numbers out of range")
    return (start-1, end)

# %% ../nbs/07_browse.ipynb 15
def _parse_browse_magic_syntax(code):
    _program_name = "temp_nbstata_syntax_name"
    run_direct_cleaned((
        f"program define {_program_name}\n"
        """ syntax [varlist(default=none)] [if] [in] [, noLabels noFormat]
            disp "%varlist%"
            foreach var in `varlist' {
                disp "`var'"
            }
            disp "%if%"
            disp `"`if'"'
            disp "%in%"
            disp `"`in'"'
            disp "%nolabels%"
            disp "`labels'"
            disp "%noformat%"
            disp "`format'"
            disp "%end%"
        end
        """), quietly=True)
    try:
        output = diverted_stata_output_quicker(f"""\
            {_program_name} {code}
            program drop {_program_name}
            """).strip()
    except Exception as e:
        run_sfi(f"capture program drop {_program_name}")
        raise(e)
    return output.replace("\n> ", "") #[c.strip() for c in var_code.split() if c] if var_code else None

# %% ../nbs/07_browse.ipynb 20
matchparts = re.compile(
            r"\A.*?"
            r"^%varlist%(?P<varlist>.*?)"
            r"%if%(?P<if>.*?)"
            r"%in%(?P<in>.*?)"
            r"%nolabels%(?P<nolabels>.*?)"
            r"%noformat%(?P<noformat>.*?)%end%", #"(\Z|---+\s*end)",
            flags=re.DOTALL + re.MULTILINE).match

# %% ../nbs/07_browse.ipynb 22
def parse_browse_magic(code):
    N = None
    m = re.match(r"\A(?P<N>[0-9]+)?[\s]*(?P<remainder>.*?)\Z", code.strip())
    if m is None:
        raise ValueError(f"syntax error: this magic must be in a code by itself on a single line")
    if m.group('N'):
        N = int(m.group('N'))
    code_minus_N = m.group('remainder')
    args = matchparts(_parse_browse_magic_syntax(code_minus_N)).groupdict()
    _var = [c.strip() for c in args['varlist'].split() if c]
    var = _var if _var else None
    if_code = args['if'].strip()
    in_code = args['in'].strip()
    nolabels = args['nolabels'].strip()
    noformat = args['noformat'].strip()
    return N, var, if_code, in_code, nolabels, noformat

# %% ../nbs/07_browse.ipynb 26
def _parse_df_params(code, count, browse=False, tail=False):
    import numpy as np
    N, var, if_code, in_code, nolabels, noformat = parse_browse_magic(code)
    sformat = not noformat
    valuelabel = not nolabels

    N_max = np.inf if browse else 5
    if N is not None:
        if browse:
            print_red("Warning: '%browse [N]' syntax is deprecated "
                      "and may be removed in v1.0.")
        N_max = N

    # Obs range
    obs_range = None
    if browse:
        start, end = in_range(in_code, count)
        if start != None and end != None:
            obs_range = range(start, end)
        elif count > N_max:
            obs_range = range(0, N_max)
    else:
        if in_code:
            print_red(f"Note: [in] not allowed for {'tail' if tail else 'head'} "
                      "magic and is ignored."
                     )
        if count > N_max:
            obs_range = range(count - N_max, count) if tail else range(0, N_max)

    return obs_range, var, if_code, valuelabel, sformat

# %% ../nbs/07_browse.ipynb 27
def get_df(obs_range, var, stata_if_code, missingval, valuelabel, sformat):
    with SelectVar(stata_if_code) as sel_varname:
        df = better_pdataframe_from_data(obs=obs_range,
                                         var=var,
                                         selectvar=sel_varname,
                                         missingval=missingval,
                                         valuelabel=valuelabel,
                                         sformat=sformat,
                                        )
        if not var and sel_varname is not None and sel_varname in df:
            df = df.drop([sel_varname], axis=1)
    return df

# %% ../nbs/07_browse.ipynb 29
def headtail_df_params(code, count, missing_config, tail=False):
    import numpy as np
    custom_missingval = missing_config != 'pandas'
    missingval = missing_config if custom_missingval else np.NaN
    obs_range, var, stata_if_code, valuelabel, sformat = (
        _parse_df_params(code, count, tail=tail)
    )
    return obs_range, var, stata_if_code, missingval, valuelabel, sformat

# %% ../nbs/07_browse.ipynb 33
def headtail_get_df(obs_range, var, stata_if_code, missingval, valuelabel, sformat):
    if not stata_if_code:
        return get_df(obs_range, var, stata_if_code, missingval, valuelabel, sformat)
    N_max = len(obs_range)
    tail = obs_range[0] != 0
    with SelectVar(stata_if_code) as sel_varname:
        df = better_pdataframe_from_data(obs=None,
                                         var=var,
                                         selectvar=sel_varname,
                                         missingval=missingval,
                                         valuelabel=valuelabel,
                                         sformat=sformat,
                                        )
        if not var and sel_varname is not None and sel_varname in df:
            df = df.drop([sel_varname], axis=1)
    return df.tail(N_max) if tail else df.head(N_max)

# %% ../nbs/07_browse.ipynb 44
def browse_df_params(code, count, missing_config):
    import numpy as np
    custom_missingval = missing_config != 'pandas'
    missingval = missing_config if custom_missingval else np.NaN
    obs_range, var, stata_if_code, valuelabel, sformat = (
        _parse_df_params(code, count, browse=True)
    )
    return obs_range, var, stata_if_code, missingval, valuelabel, sformat

# %% ../nbs/07_browse.ipynb 52
def set_ipydatagrid_height():
    from IPython.core.display import HTML
    display(HTML("<style>div.jp-Notebook .datagrid-container {min-height: 448px; }</style>"))

# %% ../nbs/07_browse.ipynb 53
def display_df_as_ipydatagrid(df):
    from ipydatagrid import DataGrid, TextRenderer
    i_renderer = TextRenderer(horizontal_alignment="right", font="11px monospace", background_color="rgb(243, 243, 243)")
    d_renderer = TextRenderer(horizontal_alignment="right", font="11px monospace")
    h_renderer = TextRenderer(horizontal_alignment="center", font="11px monospace")
    column_widths = {}
    temp_head = df.head(20)
    char_px_width = 6.05
    def name_width(name_len):
        if name_len <= 15:
            return 43 + char_px_width*name_len
        elif name_len > 22 or name_len%2 == 0:
            return 43 + char_px_width*15.3
        else:
            return 43 + char_px_width*14.3
            
    for name in list(df):
        column_widths[name] = max(
            name_width(len(name)),
            20 + char_px_width*max((len(str(value)) for value in temp_head[name])),
        )
    column_widths[" "] = 20 + char_px_width*len(str(max(df.index.values)))
    g = DataGrid(df,
                 index_name=" ",
                 editable=False,
                 selection_mode='cell',
                 base_row_size=21,
                 auto_fit_columns=False,
                 default_renderer=d_renderer,
                 header_renderer=h_renderer,
                 renderers={" ": i_renderer},
                 column_widths=column_widths,
                 layout={"height": "100%"},
                )
    g.grid_style = {
        "background_color": "rgb(255, 255, 255)",
        "header_background_color": "rgb(243, 243, 243)",
        "header_grid_line_color": "rgb(229, 229, 229)",
        "vertical_grid_line_color": "rgb(229, 229, 229)",
        "horizontal_grid_line_color": "rgb(229, 229, 229)",
        "selection_fill_color": "rgb(59, 135, 195, .25)", # 206, 225, 240
        "selection_border_color": "rgb(229, 229, 229)",
        "header_selection_fill_color": "rgb(99, 99, 99, .25)", #216
        "header_selection_border_color": "rgb(229, 229, 229)",
        "cursor_fill_color": "rgb(255, 255, 255, 0)",
        "cursor_border_color": "rgb(40, 40, 40)",
        "scroll_shadow": {'size': 0},
    }
    display(g)



"""nbstata kernel install script"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/15_install.ipynb.

# %% auto 0
__all__ = ['kernel_json', 'install_kernel_spec', 'create_conf_if_needed', 'main']

# %% ../nbs/15_install.ipynb 3
import argparse
import json
import os
import sys

from jupyter_client.kernelspec import KernelSpecManager
from IPython.utils.tempdir import TemporaryDirectory
from importlib import resources
from shutil import copyfile
from pathlib import Path
from textwrap import dedent
from fastcore.basics import IN_NOTEBOOK

# %% ../nbs/15_install.ipynb 5
kernel_json = {
    "argv": [sys.executable, "-m", "nbstata", "-f", "{connection_file}"],
    "display_name": "Stata (nbstata)",
    "language": "stata",
}

def install_kernel_spec(user=True, prefix=None):
    with TemporaryDirectory() as td:
        os.chmod(td, 0o755) # Starts off as 700, not user readable
        with open(os.path.join(td, 'kernel.json'), 'w') as f:
            json.dump(kernel_json, f, sort_keys=True)

        # Copy logo to tempdir to be installed with kernelspec
        logo_path = resources.files('nbstata').joinpath('logo-64x64.png')
        copyfile(logo_path, os.path.join(td, 'logo-64x64.png'))

        print('Installing Jupyter kernel spec')
        KernelSpecManager().install_kernel_spec(td, 'nbstata', user=user, prefix=prefix)

# %% ../nbs/15_install.ipynb 6
def _find_stata():
    # By avoiding an import of .config until we need it, we can
    # complete the installation process in virtual environments
    # without needing this submodule nor its downstream imports.
    from nbstata.config import find_dir_edition
    try:
        stata_dir, stata_ed = find_dir_edition()
    except OSError as err:
        stata_dir, stata_ed = "", ""
    return stata_dir, stata_ed

# %% ../nbs/15_install.ipynb 7
def _conf_default(stata_dir='', stata_ed=''):
    from nbstata.config import Config
    conf_default = dedent(
        f"""\
        [nbstata]
        stata_dir = {stata_dir}
        edition = {stata_ed}
        """
    )
    for key in Config.env.keys():
        if key not in {'stata_dir', 'edition'}:
            conf_default += f"{key} = {Config.env[key]}\n"
    return conf_default

# %% ../nbs/15_install.ipynb 8
def create_conf_if_needed(conf_path, conf_file_requested=True):
    """Create config file if requested or if Stata not found automatically"""
    if conf_path.is_file():
        if conf_file_requested:
            print("Configuration file already exists at:")
            print(str(conf_path))
        return
    stata_dir, stata_ed = _find_stata()
    if not stata_ed:
        msg = """\
            WARNING: Could not find Stata path.
            Please specify it manually in configuration file.
            """
        print(dedent(msg))
    if stata_ed and not conf_file_requested:
        return
    try:
        conf_dir = Path(os.path.dirname(conf_path))
        if not conf_dir.is_dir():
            os.makedirs(conf_dir)
        with conf_path.open('w') as f:
            f.write(_conf_default(stata_dir, stata_ed))
        print("Configuration file created at:")
        print(str(conf_path))
    except Exception as err:
        print(f"Attempt to create a configuration file at {str(conf_path)} failed.")
        raise(err)

# %% ../nbs/15_install.ipynb 10
def _is_root():
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False # assume not an admin on non-Unix platforms

# %% ../nbs/15_install.ipynb 12
def _conf_path(user, prefix):
    from nbstata.config import old_user_config_path, xdg_user_config_path
    if user:
        alt_conf_path = old_user_config_path()
        if alt_conf_path.is_file():
            return alt_conf_path
        else:
            return xdg_user_config_path()
    else:
        return Path(os.path.join(prefix, 'etc/nbstata.conf'))

# %% ../nbs/15_install.ipynb 14
def main(argv=None):
    ap = argparse.ArgumentParser()
    ap.add_argument('--user', action='store_true',
        help="Install to the per-user kernels registry. Default if not root.")
    ap.add_argument('--sys-prefix', action='store_true',
        help="Install to sys.prefix (e.g. a virtualenv or conda env)")
    ap.add_argument('--prefix',
        help="Install to the given prefix. "
             "Kernelspec will be installed in {PREFIX}/share/jupyter/kernels/")
    ap.add_argument('--conf-file', action='store_true',
        help="Create a configuration file.")
    args = ap.parse_args(argv)

    if args.sys_prefix:
        args.prefix = sys.prefix
    if not args.prefix and not _is_root():
        args.user = True

    install_kernel_spec(user=args.user, prefix=args.prefix)
    conf_path = _conf_path(user=args.user, prefix=sys.prefix)
    create_conf_if_needed(conf_path, conf_file_requested=args.conf_file)

# %% ../nbs/15_install.ipynb 15
#|eval: false
if __name__ == "__main__" and not IN_NOTEBOOK:
    main()
